# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import quota
from openstack import resource
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'floatingip': 1,
    'network': 2,
    'port': 3,
    'project_id': '4',
    'router': 5,
    'router_route': 30,
    'subnet': 6,
    'subnetpool': 7,
    'security_group_rule': 8,
    'security_group': 9,
    'rbac_policy': -1,
    'healthmonitor': 11,
    'listener': 12,
    'loadbalancer': 13,
    'l7policy': 14,
    'pool': 15,
    'check_limit': True,
}


class TestQuota(base.TestCase):
    def test_basic(self):
        sot = quota.Quota()
        self.assertEqual('quota', sot.resource_key)
        self.assertEqual('quotas', sot.resources_key)
        self.assertEqual('/quotas', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = quota.Quota(**EXAMPLE)
        self.assertEqual(EXAMPLE['floatingip'], sot.floating_ips)
        self.assertEqual(EXAMPLE['network'], sot.networks)
        self.assertEqual(EXAMPLE['port'], sot.ports)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['router'], sot.routers)
        self.assertEqual(EXAMPLE['router_route'], sot.router_routes)
        self.assertEqual(EXAMPLE['subnet'], sot.subnets)
        self.assertEqual(EXAMPLE['subnetpool'], sot.subnet_pools)
        self.assertEqual(
            EXAMPLE['security_group_rule'], sot.security_group_rules
        )
        self.assertEqual(EXAMPLE['security_group'], sot.security_groups)
        self.assertEqual(EXAMPLE['rbac_policy'], sot.rbac_policies)
        self.assertEqual(EXAMPLE['healthmonitor'], sot.health_monitors)
        self.assertEqual(EXAMPLE['listener'], sot.listeners)
        self.assertEqual(EXAMPLE['loadbalancer'], sot.load_balancers)
        self.assertEqual(EXAMPLE['l7policy'], sot.l7_policies)
        self.assertEqual(EXAMPLE['pool'], sot.pools)
        self.assertEqual(EXAMPLE['check_limit'], sot.check_limit)

    def test_prepare_request(self):
        body = {'id': 'ABCDEFGH', 'network': '12345'}
        quota_obj = quota.Quota(**body)
        response = quota_obj._prepare_request()
        self.assertNotIn('id', response)

    def test_alternate_id(self):
        my_project_id = 'my-tenant-id'
        body = {'project_id': my_project_id, 'network': 12345}
        quota_obj = quota.Quota(**body)
        self.assertEqual(my_project_id, resource.Resource._get_id(quota_obj))


class TestQuotaDefault(base.TestCase):
    def test_basic(self):
        sot = quota.QuotaDefault()
        self.assertEqual('quota', sot.resource_key)
        self.assertEqual('quotas', sot.resources_key)
        self.assertEqual('/quotas/%(project)s/default', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = quota.QuotaDefault(project='FAKE_PROJECT', **EXAMPLE)
        self.assertEqual(EXAMPLE['floatingip'], sot.floating_ips)
        self.assertEqual(EXAMPLE['network'], sot.networks)
        self.assertEqual(EXAMPLE['port'], sot.ports)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['router'], sot.routers)
        self.assertEqual(EXAMPLE['router_route'], sot.router_routes)
        self.assertEqual(EXAMPLE['subnet'], sot.subnets)
        self.assertEqual(EXAMPLE['subnetpool'], sot.subnet_pools)
        self.assertEqual(
            EXAMPLE['security_group_rule'], sot.security_group_rules
        )
        self.assertEqual(EXAMPLE['security_group'], sot.security_groups)
        self.assertEqual(EXAMPLE['rbac_policy'], sot.rbac_policies)
        self.assertEqual(EXAMPLE['healthmonitor'], sot.health_monitors)
        self.assertEqual(EXAMPLE['listener'], sot.listeners)
        self.assertEqual(EXAMPLE['loadbalancer'], sot.load_balancers)
        self.assertEqual(EXAMPLE['l7policy'], sot.l7_policies)
        self.assertEqual(EXAMPLE['pool'], sot.pools)
        self.assertEqual(EXAMPLE['check_limit'], sot.check_limit)
        self.assertEqual('FAKE_PROJECT', sot.project)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class AutoAllocatedTopology(resource.Resource):
    resource_name = 'auto_allocated_topology'
    resource_key = 'auto_allocated_topology'
    base_path = '/auto-allocated-topology'

    _allow_unknown_attrs_in_body = True

    # Capabilities
    allow_create = False
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_list = False

    # NOTE: this resource doesn't support list or query

    # Properties
    #: Project ID
    #: If project is not specified the topology will be created
    #: for project user is authenticated against.
    #: Will return in error if resources have not been configured correctly
    #: To use this feature auto-allocated-topology, subnet_allocation,
    #: external-net and router extensions must be enabled and set up.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)


class ValidateTopology(AutoAllocatedTopology):
    base_path = '/auto-allocated-topology/%(project)s?fields=dry-run'

    #: Validate requirements before running (Does not return topology)
    #: Will return "Deployment error:" if the resources required have not
    #: been correctly set up.
    dry_run = resource.Body('dry_run')
    project = resource.URI('project')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class BgpVpnPortAssociation(resource.Resource):
    resource_key = 'port_association'
    resources_key = 'port_associations'
    base_path = '/bgpvpn/bgpvpns/%(bgpvpn_id)s/port_associations'
    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The Id of the BGPVPN
    id = resource.Body('id')
    #: The ID of the BGPVPN who owns Network Association.
    bgpvpn_id = resource.URI('bgpvpn_id')
    #: The ID of the project that owns the BGPVPN
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The ID of a Neutron Port with which to associate the BGP VPN.
    port_id = resource.Body('port_id')
    #: Boolean flag controlling whether or not the fixed IPs of a port will be
    #: advertised to the BGPVPN (default: true).
    advertise_fixed_ips = resource.Body('advertise_fixed_ips')
    #: List of routes, each route being a dict with at least a type key,
    #: which can be prefix or bgpvpn.
    #: For the prefix type, the IP prefix (v4 or v6) to advertise is specified
    #: in the prefix key.
    #: For the bgpvpn type, the bgpvpn_id key specifies the BGPVPN from which
    #: routes will be readvertised
    routes = resource.Body('routes', type=list)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import resource


class Message(resource.Resource):
    resource_key = "messages"
    resources_key = "messages"
    base_path = "/messages"

    _query_mapping = resource.QueryParameters(
        'project_id', 'sort', 'limit', 'offset', 'marker'
    )

    # capabilities
    allow_fetch = True
    allow_delete = True
    allow_list = True

    #: Properties
    #: The date and time when the resource was created.
    #: Format is ISO 8601: CCYY-MM-DDThh:mm:ss±hh:mm
    #: Example: 2015-08-27T09:49:58-05:00
    created_at = resource.Body("created_at", type=str)
    #: The id of the event to this message.
    event_id = resource.Body("event_id", type=str)
    #: The expire time of message, message could be deleted after this time.
    guaranteed_until = resource.Body("guaranteed_until", type=str)
    #: The level of the message, possible value is only 'ERROR' now.
    message_level = resource.Body("message_level", type=str)
    #: The id of the request during which the message was created.
    request_id = resource.Body("request_id", type=str)
    #: The resource type corresponding to resource_uuid.
    resource_type = resource.Body("resource_type", type=str)
    #: The UUID of the resource during whose operation the message was created.
    resource_uuid = resource.Body("resource_uuid", type=str)
    #: The translated readable message corresponding to event_id.
    user_message = resource.Body("user_message", type=str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Callable
import functools
from typing import Any, TypeVar, cast
import warnings

from openstack import warnings as os_warnings

F = TypeVar('F', bound=Callable[..., Any])


def renamed_param(old: str, new: str) -> Callable[[F], F]:
    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if old in kwargs:
                warnings.warn(
                    f"Parameter '{old}' has been renamed to '{new}'.",
                    os_warnings.RemovedInSDK50Warning,
                    stacklevel=2,
                )
                kwargs[new] = kwargs.pop(old)
            return func(*args, **kwargs)

        return cast(F, wrapper)

    return decorator



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class ResourceProvider(resource.Resource):
    resource_key = None
    resources_key = 'resource_providers'
    base_path = '/resource_providers'

    # Capabilities

    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Filters

    _query_mapping = resource.QueryParameters(
        'name',
        'member_of',
        'resources',
        'in_tree',
        'required',
        id='uuid',
    )

    # The parent_provider_uuid and root_provider_uuid fields were introduced in
    # 1.14
    # The required query parameter was added in 1.18
    # The create operation started returning a body in 1.20
    _max_microversion = '1.20'

    # Properties

    #: Aggregates
    aggregates = resource.Body('aggregates', type=list, list_type=str)
    #: The UUID of a resource provider.
    id = resource.Body('uuid', alternate_id=True)
    #: A consistent view marker that assists with the management of concurrent
    #: resource provider updates.
    generation = resource.Body('generation')
    #: Links pertaining to this flavor. This is a list of dictionaries,
    #: each including keys ``href`` and ``rel``.
    links = resource.Body('links')
    #: The name of this resource provider.
    name = resource.Body('name')
    #: The UUID of the immediate parent of the resource provider.
    parent_provider_id = resource.Body('parent_provider_uuid')
    #: Read-only UUID of the top-most provider in this provider tree.
    root_provider_id = resource.Body('root_provider_uuid')
    #: Resource class usage counts for this resource provider.
    usages = resource.Body('usages', type=dict)

    def fetch_usages(self, session: adapter.Adapter) -> Self:
        """Fetch resource usage counts for the resource provider

        :param session: The session to use for making this request
        :return: The resource provider with usages populated
        """
        url = utils.urljoin(self.base_path, self.id, 'usages')
        microversion = self._get_microversion(session)

        response = session.get(url, microversion=microversion)
        exceptions.raise_from_response(response)
        data = response.json()

        self._body.attributes.update({'usages': data['usages']})

        return self

    def fetch_aggregates(self, session: adapter.Adapter) -> Self:
        """List aggregates set on the resource provider

        :param session: The session to use for making this request
        :return: The resource provider with aggregates populated
        """
        url = utils.urljoin(self.base_path, self.id, 'aggregates')
        microversion = self._get_microversion(session)

        response = session.get(url, microversion=microversion)
        exceptions.raise_from_response(response)
        data = response.json()

        updates = {'aggregates': data['aggregates']}
        if utils.supports_microversion(session, '1.19'):
            updates['generation'] = data['resource_provider_generation']
        self._body.attributes.update(updates)

        return self

    def set_aggregates(
        self,
        session: adapter.Adapter,
        aggregates: list[str] | None = None,
    ) -> Self:
        """Replaces aggregates on the resource provider

        :param session: The session to use for making this request
        :param list aggregates: List of aggregates
        :return: The resource provider with updated aggregates populated
        """
        url = utils.urljoin(self.base_path, self.id, 'aggregates')
        microversion = self._get_microversion(session)

        body = {
            'aggregates': aggregates or [],
        }
        if utils.supports_microversion(session, '1.19'):
            body['resource_provider_generation'] = self.generation

        response = session.put(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)
        data = response.json()

        updates = {'aggregates': data['aggregates']}
        if 'resource_provider_generation' in data:
            updates['resource_provider_generation'] = data[
                'resource_provider_generation'
            ]
        self._body.attributes.update(updates)

        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.dns.v2 import zone_nameserver
from openstack.tests.unit import base


class TestZoneNameserver(base.TestCase):
    def test_basic(self):
        sot = zone_nameserver.ZoneNameserver()
        self.assertEqual(None, sot.resource_key)
        self.assertEqual('nameservers', sot.resources_key)
        self.assertEqual('/zones/%(zone_id)s/nameservers', sot.base_path)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_commit)

        self.assertDictEqual({}, sot._query_mapping._mapping)

    def test_make_it(self):
        hostname = 'bogus-hostname'
        priority = 123

        sot = zone_nameserver.ZoneNameserver(
            hostname=hostname, priority=priority
        )
        self.assertEqual(hostname, sot.hostname)
        self.assertEqual(priority, sot.priority)



# Copyright 2024 RedHat Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.image.v2 import image_tasks
from openstack.tests.unit import base

EXAMPLE = {
    'id': '56ab5f98-2bb7-44c7-bc05-52bde37eb53b',
    'type': 'import',
    'status': 'failure',
    'owner': '2858d31bc5f54f4db66e53ab905ef566',
    'expires_at': '2024-10-10T09:28:58.000000',
    'created_at': '2024-10-08T09:28:58.000000',
    'updated_at': '2024-10-08T09:28:58.000000',
    'deleted_at': None,
    'deleted': False,
    'image_id': '56a39162-730d-401c-8a77-11bc078cf3e2',
    'request_id': 'req-7d2f073c-f6f8-4807-9fdb-5ce6b10c65c5',
    'user_id': 'dec9b6d341ec481abddf1027576c2d60',
    'input': {'image_id': '56a39162-730d-401c-8a77-11bc078cf3e2'},
    'result': None,
    'message': "Input does not contain 'import_from' field",
}


class TestImageTasks(base.TestCase):
    def test_basic(self):
        sot = image_tasks.ImageTasks()
        self.assertEqual('/images/%(image_id)s/tasks', sot.base_path)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = image_tasks.ImageTasks(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['type'], sot.type)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['owner'], sot.owner)
        self.assertEqual(EXAMPLE['expires_at'], sot.expires_at)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['deleted_at'], sot.deleted_at)
        self.assertEqual(EXAMPLE['deleted'], sot.deleted)
        self.assertEqual(EXAMPLE['image_id'], sot.image_id)
        self.assertEqual(EXAMPLE['request_id'], sot.request_id)
        self.assertEqual(EXAMPLE['user_id'], sot.user_id)
        self.assertEqual(EXAMPLE['input'], sot.input)
        self.assertEqual(EXAMPLE['result'], sot.result)
        self.assertEqual(EXAMPLE['message'], sot.message)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import volume as _volume
from openstack.tests.functional.block_storage.v3 import base


class TestVolume(base.BaseBlockStorageTest):
    def setUp(self):
        super().setUp()

        if not self.user_cloud.has_service('block-storage'):
            self.skipTest('block-storage service not supported by cloud')

        volume_name = self.getUniqueString()

        self.volume = self.block_storage_client.create_volume(
            name=volume_name,
            size=1,
        )
        self.block_storage_client.wait_for_status(
            self.volume,
            status='available',
            failures=['error'],
            interval=2,
            wait=self._wait_for_timeout,
        )
        self.assertIsInstance(self.volume, _volume.Volume)
        self.assertEqual(volume_name, self.volume.name)

    def tearDown(self):
        self.block_storage_client.delete_volume(self.volume)
        super().tearDown()

    def test_volume(self):
        # get
        volume = self.block_storage_client.get_volume(self.volume.id)
        self.assertEqual(self.volume.name, volume.name)

        # find
        volume = self.block_storage_client.find_volume(
            self.volume.name, ignore_missing=False
        )
        self.assertEqual(self.volume.id, volume.id)

        # list
        volumes = self.block_storage_client.volumes()
        # other tests may have created volumes so we don't assert that this is
        # the *only* volume present
        self.assertIn(self.volume.id, {v.id for v in volumes})

        # update
        volume_name = self.getUniqueString()
        volume_description = self.getUniqueString()
        volume = self.block_storage_client.update_volume(
            self.volume,
            name=volume_name,
            description=volume_description,
        )
        self.assertIsInstance(volume, _volume.Volume)
        volume = self.block_storage_client.get_volume(self.volume.id)
        self.assertEqual(volume_name, volume.name)
        self.assertEqual(volume_description, volume.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.tests.functional import base


class TestExtension(base.BaseFunctionalTest):
    def test_list(self):
        extensions = list(self.user_cloud.network.extensions())
        self.assertGreater(len(extensions), 0)

        for ext in extensions:
            self.assertIsInstance(ext.name, str)
            self.assertIsInstance(ext.alias, str)

    def test_find(self):
        extension = self.user_cloud.network.find_extension(
            "external-net", ignore_missing=False
        )
        self.assertEqual("Neutron external network", extension.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import builtins
from collections.abc import Generator, Mapping
from typing import Any, Literal, Self, overload
import urllib.parse

from keystoneauth1 import adapter
import requests

from openstack import exceptions
from openstack import resource


class Resource(resource.Resource):
    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None:
        """Find a resource by its name or id.

        :param session: The session to use for making this request.
        :param name_or_id: This resource's identifier, if needed by the
            request. The default is ``None``.
        :param ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the resource does not exist. When set to ``True``, None will
            be returned when attempting to find a nonexistent resource.
        :param params: Any additional parameters to be passed into underlying
            methods, such as to :meth:`~openstack.resource.Resource.existing`
            in order to pass on URI parameters.

        :return: The :class:`Resource` object matching the given name or id
            or None if nothing matches.
        :raises: :class:`openstack.exceptions.DuplicateResource` if more than
            one resource is found for this request.
        :raises: :class:`openstack.exceptions.NotFoundException` if nothing is
            found and ignore_missing is ``False``.
        """
        session = cls._get_session(session)
        # Try to short-circuit by looking directly for a matching ID.
        try:
            match = cls.existing(
                id=name_or_id,
                connection=session._get_connection(),  # type: ignore
                **params,
            )
            return match.fetch(session)
        except exceptions.SDKException:
            # DNS may return 400 when we try to do GET with name
            pass

        if (
            'name' in cls._query_mapping._mapping.keys()
            and 'name' not in params
        ):
            params['name'] = name_or_id

        data = cls.list(
            session,
            list_base_path=list_base_path,
            microversion=microversion,
            all_projects=all_projects,
            **params,
        )

        result = cls._get_one_match(name_or_id, data)
        if result is not None:
            return result

        if ignore_missing:
            return None
        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )

    @classmethod
    def list(
        cls,
        session: adapter.Adapter,
        paginated: bool = True,
        base_path: str | None = None,
        allow_unknown_params: bool = False,
        *,
        microversion: str | None = None,
        headers: dict[str, str] | None = None,
        max_items: int | None = None,
        project_id: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Generator[Self, None, None]:
        if project_id or all_projects is not None:
            if headers is None:
                headers = {}
            if project_id:
                headers["x-auth-sudo-project-id"] = str(project_id)
            if all_projects:
                headers["x-auth-all-projects"] = str(all_projects)

        return super().list(session=session, headers=headers, **params)

    @classmethod
    def _get_next_link(
        cls,
        uri: str,
        response: requests.Response,
        data: dict[str, Any],
        marker: str | None,
        limit: str | None,
        total_yielded: int,
    ) -> tuple[
        str | None, Mapping[str, str | builtins.list[str] | int | None]
    ]:
        next_link = None
        params: dict[str, list[str] | str | None] = {}
        if isinstance(data, dict):
            links = data.get('links')
            if links:
                next_link = links.get('next')

            total = data.get('metadata', {}).get('total_count')
            if total:
                # We have a kill switch
                total_count = int(total)
                if total_count <= total_yielded:
                    return None, params

        # Parse params from Link (next page URL) into params.
        # This prevents duplication of query parameters that with large
        # number of pages result in HTTP 414 error eventually.
        if next_link:
            parts = urllib.parse.urlparse(next_link)
            query_params = urllib.parse.parse_qs(parts.query)
            params.update(query_params)
            next_link = urllib.parse.urljoin(next_link, parts.path)

        # If we still have no link, and limit was given and is non-zero,
        # and the number of records yielded equals the limit, then the user
        # is playing pagination ball so we should go ahead and try once more.
        if not next_link and limit:
            next_link = uri
            params['marker'] = marker
            params['limit'] = limit

        return next_link, params



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import secret_consumer as _secret_consumer
from openstack.tests.functional import base


class TestSecretConsumer(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('key-manager')

        self.secret = self.operator_cloud.key_manager.create_secret(
            name=self.getUniqueString('secret'),
            payload='functional-secret-consumer-payload',
            payload_content_type='text/plain',
            secret_type='opaque',
        )
        self.secret_id = self.secret.secret_id
        self.assertIsNotNone(self.secret_id)

        self.addCleanup(
            self.operator_cloud.key_manager.delete_secret,
            self.secret_id,
            ignore_missing=True,
        )

    def test_secret_consumer(self):
        key_manager = self.operator_cloud.key_manager

        consumer = key_manager.create_secret_consumer(
            self.secret_id,
            service='image',
            resource_type='image',
            resource_id='123e4567-e89b-12d3-a456-426614174000',
        )
        self.assertIsInstance(consumer, _secret_consumer.SecretConsumer)

        consumers = list(key_manager.secret_consumers(self.secret_id))
        self.assertEqual(1, len(consumers))
        self.assertEqual('image', consumers[0].service)
        self.assertEqual('image', consumers[0].resource_type)

        key_manager.delete_secret_consumer(
            self.secret_id,
            service='image',
            resource_type='image',
            resource_id='123e4567-e89b-12d3-a456-426614174000',
        )

        consumers = list(key_manager.secret_consumers(self.secret_id))
        self.assertEqual(0, len(consumers))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
List resources from the Shared File System service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/shared_file_system.html
"""


def list_share_replicas(conn, **attrs):
    print("List share replicas for the shared file system:")
    for sr in conn.share.share_replicas(**attrs):
        print(sr)


def create_share_replica(conn, share_id, **attrs):
    print("Creating share replica:")
    share_replica = conn.share.create_share_replica(share_id, **attrs)
    print(share_replica)


def get_share_replica(conn, share_replica_id):
    print("Getting share replica with give Id:")
    share_replica = conn.share.share_replica(share_replica_id)
    print(share_replica)


def delete_share_replica(conn, share_replica_id):
    print("Deleting share replica with give Id:")
    conn.share.delete_share_replica(share_replica_id)


def reset_status_share_replica(conn, share_replica_id, status):
    print("Resetting status for given share replica Id:")
    conn.share.share_replica_reset_status(share_replica_id, status)


def reset_replica_state_share_replica(conn, share_replica_id, reset_state):
    print("Resetting replica state for given share replica Id:")
    conn.share.share_replica_reset_replica_state(share_replica_id, reset_state)


def force_delete_share_replica(conn, share_replica_id):
    print("Force deleting the share replica for the given Id:")
    conn.share.share_replica_force_delete(share_replica_id)


def promote_share_replica(conn, share_replica_id):
    print("Promoting share replica with given Id:")
    conn.share.share_replica_promote(share_replica_id)


def resync_share_replica(conn, share_replica_id):
    print("Resyncing share replica with given Id:")
    conn.share.share_replica_resync(share_replica_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests.functional.shared_file_system.v2 import base


class ShareSnapshotTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        self.SNAPSHOT_NAME = self.getUniqueString()
        my_share = self.operator_cloud.shared_file_system.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.operator_cloud.shared_file_system.wait_for_status(
            my_share,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        self.assertIsNotNone(my_share)
        self.assertIsNotNone(my_share.id)
        self.SHARE_ID = my_share.id

        msp = self.operator_cloud.shared_file_system.create_share_snapshot(
            share_id=self.SHARE_ID, name=self.SNAPSHOT_NAME, force=True
        )
        self.operator_cloud.shared_file_system.wait_for_status(
            msp,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        self.assertIsNotNone(msp.id)
        self.SNAPSHOT_ID = msp.id

    def tearDown(self):
        try:
            snpt = self.operator_cloud.shared_file_system.get_share_snapshot(
                self.SNAPSHOT_ID
            )
        except exceptions.ResourceNotFound:
            pass
        else:
            sot = self.operator_cloud.shared_file_system.delete_share_snapshot(
                snpt, ignore_missing=False
            )
            self.operator_cloud.shared_file_system.wait_for_delete(
                snpt, interval=2, wait=self._wait_for_timeout
            )
            self.assertIsNone(sot)

        sot = self.operator_cloud.shared_file_system.delete_share(
            self.SHARE_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_get(self):
        sot = self.operator_cloud.shared_file_system.get_share_snapshot(
            self.SNAPSHOT_ID
        )
        self.assertEqual(self.SNAPSHOT_NAME, sot.name)

    def test_list(self):
        snaps = self.operator_cloud.shared_file_system.share_snapshots(
            details=True
        )
        self.assertGreater(len(list(snaps)), 0)
        for snap in snaps:
            for attribute in (
                'id',
                'name',
                'created_at',
                'updated_at',
                'description',
                'share_id',
                'share_proto',
                'share_size',
                'size',
                'status',
                'user_id',
            ):
                self.assertTrue(hasattr(snap, attribute))

    def test_update(self):
        u_snap = self.operator_cloud.shared_file_system.update_share_snapshot(
            self.SNAPSHOT_ID, display_description='updated share snapshot'
        )
        get_u_snap = self.operator_cloud.shared_file_system.get_share_snapshot(
            u_snap.id
        )
        self.assertEqual('updated share snapshot', get_u_snap.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Literal, Self, overload

from keystoneauth1 import adapter

from openstack import exceptions
from openstack.image import _download
from openstack import resource


class Image(resource.Resource, _download.DownloadMixin):
    resource_key = 'image'
    resources_key = 'images'
    base_path = '/images'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Store all unknown attributes under 'properties' in the object.
    # Remotely they would be still in the resource root
    _store_unknown_attrs_as_properties = True

    _query_mapping = resource.QueryParameters(
        'name',
        'container_format',
        'disk_format',
        'status',
        'size_min',
        'size_max',
    )

    #: Hash of the image data used. The Image service uses this value
    #: for verification.
    checksum = resource.Body('checksum')
    #: The container format refers to whether the VM image is in a file
    #: format that also contains metadata about the actual VM.
    #: Container formats include OVF and Amazon AMI. In addition,
    #: a VM image might not have a container format - instead,
    #: the image is just a blob of unstructured data.
    container_format = resource.Body('container_format')
    #: A URL to copy an image from
    copy_from = resource.Body('copy_from')
    #: The timestamp when this image was created.
    created_at = resource.Body('created_at')
    #: Valid values are: aki, ari, ami, raw, iso, vhd, vdi, qcow2, or vmdk.
    #: The disk format of a VM image is the format of the underlying
    #: disk image. Virtual appliance vendors have different formats for
    #: laying out the information contained in a VM disk image.
    disk_format = resource.Body('disk_format')
    #: Defines whether the image can be deleted.
    #: *Type: bool*
    is_protected = resource.Body('protected', type=bool)
    #: ``True`` if this is a public image.
    #: *Type: bool*
    is_public = resource.Body('is_public', type=bool)
    #: A location for the image identified by a URI
    location = resource.Body('location')
    #: The minimum disk size in GB that is required to boot the image.
    min_disk = resource.Body('min_disk')
    #: The minimum amount of RAM in MB that is required to boot the image.
    min_ram = resource.Body('min_ram')
    #: Name for the image. Note that the name of an image is not unique
    #: to a Glance node. The API cannot expect users to know the names
    #: of images owned by others.
    name = resource.Body('name')
    #: The ID of the owner, or project, of the image.
    owner = resource.Body('owner', alias='owner_id')
    #: The ID of the owner, or project, of the image. (backwards compat)
    owner_id = resource.Body('owner', alias='owner')
    #: Properties, if any, that are associated with the image.
    properties = resource.Body('properties')
    #: The size of the image data, in bytes.
    size = resource.Body('size')
    #: The image status.
    status = resource.Body('status')
    #: The timestamp when this image was last updated.
    updated_at = resource.Body('updated_at')

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None:
        """Find a resource by its name or id.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param name_or_id: This resource's identifier, if needed by
                           the request. The default is ``None``.
        :param bool ignore_missing: When set to ``False``
                    :class:`~openstack.exceptions.NotFoundException` will be
                    raised when the resource does not exist.
                    When set to ``True``, None will be returned when
                    attempting to find a nonexistent resource.
        :param dict params: Any additional parameters to be passed into
                            underlying methods, such as to
                            :meth:`~openstack.resource.Resource.existing`
                            in order to pass on URI parameters.

        :return: The :class:`Resource` object matching the given name or id
                 or None if nothing matches.
        :raises: :class:`openstack.exceptions.DuplicateResource` if more
                 than one resource is found for this request.
        :raises: :class:`openstack.exceptions.NotFoundException` if nothing
                 is found and ignore_missing is ``False``.
        """
        session = cls._get_session(session)
        # Try to short-circuit by looking directly for a matching ID.
        try:
            match = cls.existing(
                id=name_or_id,
                connection=session._get_connection(),  # type: ignore
                **params,
            )
            return match.fetch(session, **params)
        except exceptions.NotFoundException:
            pass

        params['name'] = name_or_id

        data = cls.list(
            session,
            base_path='/images/detail',
            all_projects=all_projects,
            **params,
        )

        result = cls._get_one_match(name_or_id, data)
        if result is not None:
            return result

        if ignore_missing:
            return None
        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import _base
from openstack import resource


class Limit(_base.Resource):
    """DNS Limit Resource"""

    resource_key = 'limit'
    base_path = '/limits'

    # capabilities
    allow_list = True

    #: Properties
    #: The max amount of items allowed per page
    max_page_limit = resource.Body('max_page_limit', type=int)
    #: The max length of a recordset name
    max_recordset_name_length = resource.Body(
        'max_recordset_name_length', type=int
    )
    #: The max amount of records contained in a recordset
    max_recordset_records = resource.Body('max_recordset_records', type=int)
    #: The max length of a zone name
    max_zone_name_length = resource.Body('max_zone_name_length', type=int)
    #: The max amount of records in a zone
    max_zone_records = resource.Body('max_zone_records', type=int)
    #: The max amount of recordsets per zone
    max_zone_recordsets = resource.Body('max_zone_recordsets', type=int)
    #: The max amount of zones for this project
    max_zones = resource.Body('max_zones', type=int)
    #: The lowest ttl allowed on this system
    min_ttl = resource.Body('min_ttl', type=int)



# Apache 2 header omitted for brevity

from openstack import service_description
from openstack.fake.v2 import _proxy as _proxy_v2


class FakeService(service_description.ServiceDescription):
    """The fake service."""

    supported_versions = {
        '2': _proxy_v2.Proxy,
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import qos_spec
from openstack.tests.unit import base

QOS_SPEC = {
    "name": "reliability-spec",
}


class TestSnapshot(base.TestCase):
    def test_basic(self):
        sot = qos_spec.QoSSpec(**QOS_SPEC)
        self.assertEqual("qos_specs", sot.resource_key)
        self.assertEqual("qos_specs", sot.resources_key)
        self.assertEqual("/qos-specs", sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                "project_id": "project_id",
                "limit": "limit",
                "offset": "offset",
                "marker": "marker",
                "sort_dir": "sort_dir",
                "sort_key": "sort_key",
                "sort": "sort",
            },
            sot._query_mapping._mapping,
        )

    def test_create_basic(self):
        sot = qos_spec.QoSSpec(**QOS_SPEC)
        self.assertEqual(QOS_SPEC["name"], sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import _base
from openstack import resource


class TLD(_base.Resource):
    """DNS TLD Resource"""

    resources_key = "tlds"
    base_path = "/tlds"

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    commit_method = "PATCH"

    _query_mapping = resource.QueryParameters(
        "name",
        "description",
        "limit",
        "marker",
    )

    #: TLD name
    name = resource.Body("name")
    #: TLD description
    description = resource.Body("description")
    #: Timestamp when the tld was created
    created_at = resource.Body("created_at")
    #: Timestamp when the tld was last updated
    updated_at = resource.Body("updated_at")
    #: Links contains a `self` pertaining to this tld or a `next` pertaining
    #: to next page
    links = resource.Body("links", type=dict)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self, cast

from collections.abc import Mapping
from keystoneauth1 import adapter
import requests

from openstack._utils import renamed_param
from openstack.common import metadata
from openstack.common import tag
from openstack.compute.v2 import flavor as _flavor
from openstack.compute.v2 import volume_attachment as _volume_attachment
from openstack import exceptions
from openstack.image.v2 import image as _image
from openstack import resource
from openstack import types
from openstack import utils


CONSOLE_TYPE_ACTION_MAPPING = {
    'novnc': 'os-getVNCConsole',
    'xvpvnc': 'os-getVNCConsole',
    'spice-html5': 'os-getSPICEConsole',
    'spice-direct': 'os-getSPICEConsole',
    'rdp-html5': 'os-getRDPConsole',
    'serial': 'os-getSerialConsole',
}


class Server(resource.Resource, metadata.MetadataMixin, tag.TagMixin):
    resource_key = 'server'
    resources_key = 'servers'
    base_path = '/servers'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        "auto_disk_config",
        "availability_zone",
        "created_at",
        "description",
        "flavor",
        "hostname",
        "image",
        "kernel_id",
        "key_name",
        "launch_index",
        "launched_at",
        "locked",
        "locked_by",
        "name",
        "node",
        "power_state",
        "progress",
        "project_id",
        "ramdisk_id",
        "reservation_id",
        "root_device_name",
        "status",
        "task_state",
        "terminated_at",
        "user_id",
        "vm_state",
        "sort_key",
        "sort_dir",
        "pinned_availability_zone",
        access_ipv4="access_ip_v4",
        access_ipv6="access_ip_v6",
        has_config_drive="config_drive",
        deleted_only="deleted",
        compute_host="host",
        is_soft_deleted="soft_deleted",
        ipv4_address="ip",
        ipv6_address="ip6",
        changes_since="changes-since",
        changes_before="changes-before",
        id="uuid",
        all_projects="all_tenants",
        **tag.TagMixin._tag_query_parameters,
    )

    _max_microversion = '2.100'

    #: A list of dictionaries holding links relevant to this server.
    links = resource.Body('links')

    access_ipv4 = resource.Body('accessIPv4')
    access_ipv6 = resource.Body('accessIPv6')
    #: A dictionary of addresses this server can be accessed through.
    #: The dictionary contains keys such as ``private`` and ``public``,
    #: each containing a list of dictionaries for addresses of that type.
    #: The addresses are contained in a dictionary with keys ``addr``
    #: and ``version``, which is either 4 or 6 depending on the protocol
    #: of the IP address. *Type: dict*
    addresses = resource.Body('addresses', type=dict)
    #: When a server is first created, it provides the administrator password.
    admin_password = resource.Body('adminPass')
    #: A list of an attached volumes. Each item in the list contains at least
    #: an "id" key to identify the specific volumes.
    attached_volumes = resource.Body(
        'os-extended-volumes:volumes_attached',
        aka='volumes',
        type=list,
        list_type=_volume_attachment.VolumeAttachment,
        default=[],
    )
    #: The name of the availability zone this server is a part of.
    availability_zone = resource.Body('OS-EXT-AZ:availability_zone')
    #: Enables fine grained control of the block device mapping for an
    #: instance. This is typically used for booting servers from volumes.
    block_device_mapping = resource.Body('block_device_mapping_v2')
    #: Indicates whether or not a config drive was used for this server.
    config_drive = resource.Body('config_drive')
    #: The name of the compute host on which this instance is running.
    #: Appears in the response for administrative users only.
    compute_host = resource.Body('OS-EXT-SRV-ATTR:host')
    #: Timestamp of when the server was created.
    created_at = resource.Body('created')
    #: The description of the server. Before microversion
    #: 2.19 this was set to the server name.
    description = resource.Body('description')
    #: The disk configuration. Either AUTO or MANUAL.
    disk_config = resource.Body('OS-DCF:diskConfig')
    #: The flavor reference, as a ID or full URL, for the flavor to use for
    #: this server.
    flavor_id = resource.Body('flavorRef')
    #: The flavor property as returned from server.
    flavor = resource.Body('flavor', type=_flavor.Flavor)
    #: Indicates whether a configuration drive enables metadata injection.
    #: Not all cloud providers enable this feature.
    has_config_drive = resource.Body('config_drive')
    #: An ID representing the host of this server.
    host_id = resource.Body('hostId')
    #: A fault object. Only available when the server status
    #: is ERROR or DELETED and a fault occurred.
    fault = resource.Body('fault')
    #: The host to boot the server on.
    host = resource.Body('host')
    #: The host status.
    host_status = resource.Body('host_status')
    #: The hostname set on the instance when it is booted.
    #: By default, it appears in the response for administrative users only.
    hostname = resource.Body('OS-EXT-SRV-ATTR:hostname')
    #: The hypervisor host name. Appears in the response for administrative
    #: users only.
    hypervisor_hostname = resource.Body('OS-EXT-SRV-ATTR:hypervisor_hostname')
    #: The image reference, as a ID or full URL, for the image to use for
    #: this server.
    image_id = resource.Body('imageRef')
    #: The image property as returned from server.
    image = resource.Body('image', type=_image.Image)
    #: The instance name. The Compute API generates the instance name from the
    #: instance name template. Appears in the response for administrative users
    #: only.
    instance_name = resource.Body('OS-EXT-SRV-ATTR:instance_name')
    #: The address to use to connect to this server from the current calling
    #: context. This will be set to public_ipv6 if the calling host has
    #: routable ipv6 addresses, and to private_ipv4 if the Connection was
    #: created with private=True. Otherwise it will be set to public_ipv4.
    interface_ip = resource.Computed('interface_ip', default='')
    # The locked status of the server
    is_locked = resource.Body('locked', type=bool)
    #: The UUID of the kernel image when using an AMI. Will be null if not.
    #: By default, it appears in the response for administrative users only.
    kernel_id = resource.Body('OS-EXT-SRV-ATTR:kernel_id')
    #: The name of an associated keypair
    key_name = resource.Body('key_name')
    #: When servers are launched via multiple create, this is the
    #: sequence in which the servers were launched. By default, it
    #: appears in the response for administrative users only.
    launch_index = resource.Body('OS-EXT-SRV-ATTR:launch_index', type=int)
    #: The timestamp when the server was launched.
    launched_at = resource.Body('OS-SRV-USG:launched_at')
    #: The reason the server was locked, if any.
    locked_reason = resource.Body('locked_reason')
    #: The maximum number of servers to create.
    max_count = resource.Body('max_count')
    #: The minimum number of servers to create.
    min_count = resource.Body('min_count')
    #: A networks object. Required parameter when there are multiple
    #: networks defined for the tenant. When you do not specify the
    #: networks parameter, the server attaches to the only network
    #: created for the current tenant.
    networks = resource.Body('networks')
    #: Personality files. This should be a list of dicts with each dict
    #: containing a file name ('name') and a base64-encoded file contents
    #: ('contents')
    personality = resource.Body('personality', type=list)
    #: The availability zone requested during server creation OR pinned
    #: availability zone, which is configured using default_schedule_zone
    #: config option.
    pinned_availability_zone = resource.Body('pinned_availability_zone')
    #: The power state of this server.
    power_state = resource.Body('OS-EXT-STS:power_state')
    #: While the server is building, this value represents the percentage
    #: of completion. Once it is completed, it will be 100.  *Type: int*
    progress = resource.Body('progress', type=int)
    #: The ID of the project this server is associated with.
    project_id = resource.Body('tenant_id')

    #: The private IPv4 address of this server
    private_v4 = resource.Computed('private_v4', default='')
    #: The private IPv6 address of this server
    private_v6 = resource.Computed('private_v6', default='')

    #: The public IPv4 address of this server
    public_v4 = resource.Computed('public_v4', default='')
    #: The public IPv6 address of this server
    public_v6 = resource.Computed('public_v6', default='')

    #: The UUID of the ramdisk image when using an AMI. Will be null if not.
    #: By default, it appears in the response for administrative users only.
    ramdisk_id = resource.Body('OS-EXT-SRV-ATTR:ramdisk_id')
    #: The reservation id for the server. This is an id that can be
    #: useful in tracking groups of servers created with multiple create,
    #: that will all have the same reservation_id. By default, it appears
    #: in the response for administrative users only.
    reservation_id = resource.Body('OS-EXT-SRV-ATTR:reservation_id')
    #: The root device name for the instance By default, it appears in the
    #: response for administrative users only.
    root_device_name = resource.Body('OS-EXT-SRV-ATTR:root_device_name')
    #: The dictionary of data to send to the scheduler.
    scheduler_hints = resource.Body('OS-SCH-HNT:scheduler_hints', type=dict)
    #: A list of applicable security groups. Each group contains keys for
    #: description, name, id, and rules.
    security_groups = resource.Body(
        'security_groups', type=list, list_type=dict
    )
    #: The UUIDs of the server groups to which the server belongs.
    #: Currently this can contain at most one entry.
    server_groups = resource.Body('server_groups', type=list)
    #: The state this server is in. Valid values include ``ACTIVE``,
    #: ``BUILDING``, ``DELETED``, ``ERROR``, ``HARD_REBOOT``, ``PASSWORD``,
    #: ``PAUSED``, ``REBOOT``, ``REBUILD``, ``RESCUED``, ``RESIZED``,
    #: ``REVERT_RESIZE``, ``SHUTOFF``, ``SOFT_DELETED``, ``STOPPED``,
    #: ``SUSPENDED``, ``UNKNOWN``, or ``VERIFY_RESIZE``.
    status = resource.Body('status')
    #: The task state of this server.
    task_state = resource.Body('OS-EXT-STS:task_state')
    #: The timestamp when the server was terminated (if it has been).
    terminated_at = resource.Body('OS-SRV-USG:terminated_at')
    #: A list of trusted certificate IDs, that were used during image
    #: signature verification to verify the signing certificate.
    trusted_image_certificates = resource.Body(
        'trusted_image_certificates', type=list
    )
    #: Timestamp of when this server was last updated.
    updated_at = resource.Body('updated')
    #: Configuration information or scripts to use upon launch.
    #: Must be Base64 encoded.
    user_data = resource.Body('OS-EXT-SRV-ATTR:user_data')
    #: The ID of the owners of this server.
    user_id = resource.Body('user_id')
    #: The VM state of this server.
    vm_state = resource.Body('OS-EXT-STS:vm_state')

    def _prepare_request(
        self,
        requires_id: bool | None = True,
        prepend_key: bool = True,
        patch: bool = False,
        base_path: str | None = None,
        params: Any = None,
        **kwargs: Any,
    ) -> resource._Request:
        request = super()._prepare_request(
            requires_id=requires_id,
            prepend_key=prepend_key,
            patch=patch,
            base_path=base_path,
            params=params,
            **kwargs,
        )

        server_body = cast(dict[str, Any], request.body)[self.resource_key]

        # Some names exist without prefix on requests but with a prefix
        # on responses. If we find that we've populated one of these
        # attributes with something and then go to make a request, swap out
        # the name to the bare version.

        # Availability Zones exist with a prefix on response, but not request
        az_key = "OS-EXT-AZ:availability_zone"
        if az_key in server_body:
            server_body["availability_zone"] = server_body.pop(az_key)

        # User Data exists with a prefix on response, but not request
        ud_key = "OS-EXT-SRV-ATTR:user_data"
        if ud_key in server_body:
            server_body["user_data"] = server_body.pop(ud_key)

        # Scheduler hints are sent in a top-level scope, not within the
        # resource_key scope like everything else. If we try to send
        # scheduler_hints, pop them out of the resource_key scope and into
        # their own top-level scope.
        hint_key = "OS-SCH-HNT:scheduler_hints"
        if hint_key in server_body:
            cast(dict[str, Any], request.body)[hint_key] = server_body.pop(
                hint_key
            )

        # hostname exists with a prefix on response, but not request
        hostname_key = "OS-EXT-SRV-ATTR:hostname"
        if hostname_key in server_body:
            server_body["hostname"] = server_body.pop(hostname_key)

        return request

    def _action(
        self,
        session: adapter.Adapter,
        body: Mapping[str, object],
        microversion: str | None = None,
    ) -> requests.Response:
        """Preform server actions given the message body."""
        # NOTE: This is using Server.base_path instead of self.base_path
        # as both Server and ServerDetail instances can be acted on, but
        # the URL used is sans any additional /detail/ part.
        url = utils.urljoin(Server.base_path, self.id, 'action')
        headers = {'Accept': ''}

        # these aren't all necessary "commit" actions (i.e. updates) but it's
        # good enough...
        if microversion is None:
            microversion = self._get_microversion(session)

        response = session.post(
            url,
            json=body,
            headers=headers,
            microversion=microversion,
        )
        exceptions.raise_from_response(response)
        return response

    def change_password(
        self,
        session: adapter.Adapter,
        password: str,
        *,
        microversion: str | None = None,
    ) -> None:
        """Change the administrator password to the given password.

        :param session: The session to use for making this request.
        :param password: The new password.
        :returns: None
        """
        body = {'changePassword': {'adminPass': password}}
        self._action(session, body, microversion=microversion)

    def get_password(
        self, session: adapter.Adapter, *, microversion: str | None = None
    ) -> str | None:
        """Get the encrypted administrator password.

        :param session: The session to use for making this request.
        :returns: The encrypted administrator password.
        """
        url = utils.urljoin(Server.base_path, self.id, 'os-server-password')
        if microversion is None:
            microversion = self._get_microversion(session)

        response = session.get(url, microversion=microversion)
        exceptions.raise_from_response(response)

        data = response.json()
        return cast(str | None, data.get('password'))

    def clear_password(
        self, session: adapter.Adapter, *, microversion: str | None = None
    ) -> None:
        """Clear the administrator password.

        This removes the password from the database. It does not actually
        change the server password.

        :param session: The session to use for making this request.
        :returns: None
        """
        url = utils.urljoin(Server.base_path, self.id, 'os-server-password')
        if microversion is None:
            microversion = self._get_microversion(session)

        response = session.delete(url, microversion=microversion)
        exceptions.raise_from_response(response)

    def reboot(self, session: adapter.Adapter, reboot_type: str) -> None:
        """Reboot server where reboot_type might be 'SOFT' or 'HARD'.

        :param session: The session to use for making this request.
        :param reboot_type: The type of reboot. One of: ``SOFT``, ``HARD``.
        :returns: None
        """
        body = {'reboot': {'type': reboot_type}}
        self._action(session, body)

    def force_delete(self, session: adapter.Adapter) -> None:
        """Force delete the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {'forceDelete': None}
        self._action(session, body)

    def rebuild(
        self,
        session: adapter.Adapter,
        image: str | _image.Image,
        name: str | None | types.Unset = types.UNSET,
        admin_password: str | None | types.Unset = types.UNSET,
        preserve_ephemeral: bool | types.Unset = types.UNSET,
        access_ipv4: str | None | types.Unset = types.UNSET,
        access_ipv6: str | None | types.Unset = types.UNSET,
        metadata: dict[str, object] | None | types.Unset = types.UNSET,
        user_data: dict[str, object] | None | types.Unset = types.UNSET,
        key_name: str | None | types.Unset = types.UNSET,
        description: str | None | types.Unset = types.UNSET,
        trusted_image_certificates: list[str]
        | None
        | types.Unset = types.UNSET,
        hostname: str | None | types.Unset = types.UNSET,
    ) -> Self:
        """Rebuild the server with the given arguments.

        :param session: The session to use for making this request.
        :param image: The image to rebuild to. Either an ID or a
            :class:`~openstack.image.v2.image.Image` instance.
        :param name: A name to set on the rebuilt server. (Optional)
        :param admin_password: An admin password to set on the rebuilt server.
            (Optional)
        :param preserve_ephemeral: Whether to preserve the ephemeral drive
            during the rebuild. (Optional)
        :param access_ipv4: An IPv4 address that will be used to access the
            rebuilt server. (Optional)
        :param access_ipv6: An IPv6 address that will be used to access the
            rebuilt server. (Optional)
        :param metadata: Metadata to set on the updated server. (Optional)
        :param user_data: User data to set on the updated server. (Optional)
        :param key_name: A key name to set on the updated server. (Optional)
        :param description: The description to set on the updated server.
            (Optional) (Requires API microversion 2.19)
        :param trusted_image_certificates: The trusted image certificates to
            set on the updated server. (Optional) (Requires API microversion
            2.78)
        :param hostname: The hostname to set on the updated server. (Optional)
            (Requires API microversion 2.90)
        :returns: The updated server.
        """
        action: dict[str, Any] = {'imageRef': resource.Resource._get_id(image)}
        if preserve_ephemeral is not types.UNSET:
            action['preserve_ephemeral'] = preserve_ephemeral
        if name is not types.UNSET:
            action['name'] = name
        if admin_password is not types.UNSET:
            action['adminPass'] = admin_password
        if access_ipv4 is not types.UNSET:
            action['accessIPv4'] = access_ipv4
        if access_ipv6 is not types.UNSET:
            action['accessIPv6'] = access_ipv6
        if metadata is not types.UNSET:
            action['metadata'] = metadata
        if user_data is not types.UNSET:
            action['user_data'] = user_data
        if key_name is not types.UNSET:
            action['key_name'] = key_name
        if description is not types.UNSET:
            action['description'] = description
        if trusted_image_certificates is not types.UNSET:
            action['trusted_image_certificates'] = trusted_image_certificates
        if hostname is not types.UNSET:
            action['hostname'] = hostname

        body = {'rebuild': action}
        response = self._action(session, body)
        self._translate_response(response)
        return self

    def resize(self, session: adapter.Adapter, flavor: Any) -> None:
        """Resize server to flavor reference.

        :param session: The session to use for making this request.
        :param flavor: The server to resize to.
        :returns: None
        """
        body = {'resize': {'flavorRef': flavor}}
        self._action(session, body)

    def confirm_resize(self, session: adapter.Adapter) -> None:
        """Confirm the resize of the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {'confirmResize': None}
        self._action(session, body)

    def revert_resize(self, session: adapter.Adapter) -> None:
        """Revert the resize of the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {'revertResize': None}
        self._action(session, body)

    def create_image(
        self,
        session: adapter.Adapter,
        name: str,
        metadata: dict[str, str] | None = None,
    ) -> str:
        """Create image from server.

        :param session: The session to use for making this request.
        :param name: The name to use for the created image.
        :param metadata: Metadata to set on the created image. (Optional)
        :returns: image uuid
        """
        action: dict[str, object] = {'name': name}
        if metadata is not None:
            action['metadata'] = metadata
        body = {'createImage': action}

        # Microversion 2.45 changed the createImage response to return the
        # image ID in a JSON body instead of a Location header.
        microversion = None
        if utils.supports_microversion(session, '2.45'):
            microversion = '2.45'
        response = self._action(session, body, microversion)

        try:
            # With microversion 2.45+ the image ID is in the response body;
            # for older microversions it is in the Location header.
            response_body = response.json()
        except Exception:
            response_body = None
        if response_body and 'image_id' in response_body:
            image_id: str = response_body['image_id']
        else:
            image_id = response.headers['Location'].rsplit('/', 1)[1]

        return image_id

    def add_security_group(
        self, session: adapter.Adapter, security_group_name: str
    ) -> None:
        """Add a security group to the server.

        :param session: The session to use for making this request.
        :param security_group_name: The security group to add to the server.
        :returns: None
        """
        body = {"addSecurityGroup": {"name": security_group_name}}
        self._action(session, body)

    def remove_security_group(
        self, session: adapter.Adapter, security_group_name: str
    ) -> None:
        """Remove a security group from the server.

        :param session: The session to use for making this request.
        :param security_group_name: The security group to remove from the
            server.
        :returns: None
        """
        body = {"removeSecurityGroup": {"name": security_group_name}}
        self._action(session, body)

    def reset_state(self, session: adapter.Adapter, state: str) -> None:
        """Reset server state.

        This is admin-only by default.

        :param session: The session to use for making this request.
        :param state: The state to set on the server.
        :returns: None
        """
        body = {"os-resetState": {"state": state}}
        self._action(session, body)

    def add_fixed_ip(self, session: adapter.Adapter, network_id: str) -> None:
        """Add a fixed IP to the server.

        This is effectively an alias for adding a network.

        :param session: The session to use for making this request.
        :param network_id: The network to connect the server to.
        :returns: None
        """
        body = {"addFixedIp": {"networkId": network_id}}
        self._action(session, body)

    def remove_fixed_ip(self, session: adapter.Adapter, address: str) -> None:
        """Remove a fixed IP from the server.

        This is effectively an alias from removing a port from the server.

        :param session: The session to use for making this request.
        :param network_id: The address to remove from the server.
        :returns: None
        """
        body = {"removeFixedIp": {"address": address}}
        self._action(session, body)

    def add_floating_ip(
        self,
        session: adapter.Adapter,
        address: str,
        fixed_address: str | None = None,
    ) -> None:
        """Add a floating IP to the server.

        :param session: The session to use for making this request.
        :param address: The floating IP address to associate with the server.
        :param fixed_address: A fixed IP address with which to associated the
            floating IP. (Optional)
        :returns: None
        """
        body = {"addFloatingIp": {"address": address}}
        if fixed_address is not None:
            body['addFloatingIp']['fixed_address'] = fixed_address
        self._action(session, body)

    def remove_floating_ip(
        self, session: adapter.Adapter, address: str
    ) -> None:
        """Remove a floating IP from the server.

        :param session: The session to use for making this request.
        :param address: The floating IP address to disassociate from the
            server.
        :returns: None
        """
        body = {"removeFloatingIp": {"address": address}}
        self._action(session, body)

    def backup(
        self,
        session: adapter.Adapter,
        name: str,
        backup_type: str,
        rotation: int,
    ) -> str:
        """Create a backup of the server.

        :param session: The session to use for making this request.
        :param name: The name to use for the backup image.
        :param backup_type: The type of backup. The value and meaning of this
            atribute is user-defined and can be used to separate backups of
            different types. For example, this could be used to distinguish
            between ``daily`` and ``weekly`` backups.
        :param rotation: The number of backups to retain. All images older than
            the rotation'th image will be deleted.
        :returns: image uuid
        """
        body = {
            "createBackup": {
                "name": name,
                "backup_type": backup_type,
                "rotation": rotation,
            }
        }
        # Microversion 2.45 changed the createBackup response to return the
        # image ID in a JSON body instead of a Location header.
        microversion = None
        if utils.supports_microversion(session, '2.45'):
            microversion = '2.45'
        response = self._action(session, body, microversion)

        try:
            # With microversion 2.45+ the image ID is in the response body;
            # for older microversions it is in the Location header.
            response_body = response.json()
        except Exception:
            response_body = None
        if response_body and 'image_id' in response_body:
            image_id: str = response_body['image_id']
        else:
            image_id = response.headers['Location'].rsplit('/', 1)[1]

        return image_id

    def pause(self, session: adapter.Adapter) -> None:
        """Pause the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"pause": None}
        self._action(session, body)

    def unpause(self, session: adapter.Adapter) -> None:
        """Unpause the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"unpause": None}
        self._action(session, body)

    def suspend(self, session: adapter.Adapter) -> None:
        """Suspend the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"suspend": None}
        self._action(session, body)

    def resume(self, session: adapter.Adapter) -> None:
        """Resume the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"resume": None}
        self._action(session, body)

    def lock(
        self, session: adapter.Adapter, locked_reason: str | None = None
    ) -> None:
        """Lock the server.

        :param session: The session to use for making this request.
        :param locked_reason: The reason for locking the server.
        :returns: None
        """
        body: dict[str, object] = {"lock": None}
        if locked_reason is not None:
            body["lock"] = {
                "locked_reason": locked_reason,
            }
        self._action(session, body)

    def unlock(self, session: adapter.Adapter) -> None:
        """Unlock the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"unlock": None}
        self._action(session, body)

    @renamed_param('image_ref', 'image')
    def rescue(
        self,
        session: adapter.Adapter,
        admin_pass: str | None = None,
        image: str | _image.Image | None = None,
    ) -> None:
        """Rescue the server.

        This is admin-only by default.

        :param session: The session to use for making this request.
        :param admin_pass: A new admin password to set on the rescued server.
            (Optional)
        :param image: The image to use when rescuing the server. If not
            provided, the server will use the existing image. (Optional)
        :returns: None
        """
        body: dict[str, Any] = {"rescue": {}}
        if admin_pass is not None:
            body["rescue"]["adminPass"] = admin_pass
        if image is not None:
            image_id = resource.Resource._get_id(image)
            body["rescue"]["rescue_image_ref"] = image_id
        self._action(session, body)

    def unrescue(self, session: adapter.Adapter) -> None:
        """Unrescue the server.

        This is admin-only by default.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"unrescue": None}
        self._action(session, body)

    def evacuate(
        self,
        session: adapter.Adapter,
        host: str | None = None,
        admin_pass: str | None = None,
        force: bool | None = None,
        on_shared_storage: bool | None = None,
    ) -> None:
        """Evacuate the server.

        :param session: The session to use for making this request.
        :param host: The host to evacuate the instance to. (Optional)
        :param admin_pass: The admin password to set on the evacuated instance.
            (Optional)
        :param force: Whether to force evacuation.
        :param on_shared_storage: Whether the host is using shared storage.
            (Optional) (Only supported before microversion 2.14)
        :returns: None
        """
        body: dict[str, Any] = {"evacuate": {}}
        if host is not None:
            body["evacuate"]["host"] = host
        if admin_pass is not None:
            body["evacuate"]["adminPass"] = admin_pass
        if force is not None:
            body["evacuate"]["force"] = force
        if on_shared_storage is not None:
            body["evacuate"]["onSharedStorage"] = on_shared_storage
        self._action(session, body)

    def start(self, session: adapter.Adapter) -> None:
        """Start the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"os-start": None}
        self._action(session, body)

    def stop(self, session: adapter.Adapter) -> None:
        """Stop the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"os-stop": None}
        self._action(session, body)

    def restore(self, session: adapter.Adapter) -> None:
        """Restore the server.

        This is only supported if the server is soft-deleted. This is
        cloud-specific.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"restore": None}
        self._action(session, body)

    def shelve(self, session: adapter.Adapter) -> None:
        """Shelve the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"shelve": None}
        self._action(session, body)

    def shelve_offload(self, session: adapter.Adapter) -> None:
        """Shelve-offload the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"shelveOffload": None}
        self._action(session, body)

    def unshelve(
        self,
        session: adapter.Adapter,
        availability_zone: str | None | types.Unset = types.UNSET,
        host: str | None = None,
    ) -> None:
        """Unshelve the server.

        :param session: The session to use for making this request.
        :param availability_zone: If specified the instance will be unshelved
            to the availability_zone.
            If None is passed the instance defined availability_zone is unpin
            and the instance will be scheduled to any availability_zone (free
            scheduling).
            If not specified the instance will be unshelved to either its
            defined availability_zone or any availability_zone
            (free scheduling).
        :param host: If specified the host to unshelve the instance.
        """
        data: dict[str, object] = {}
        if host:
            data["host"] = host
        if availability_zone is None or isinstance(availability_zone, str):
            data["availability_zone"] = availability_zone
        body = {'unshelve': data or None}
        self._action(session, body)

    def migrate(
        self, session: adapter.Adapter, *, host: str | None = None
    ) -> None:
        """Migrate the server.

        :param session: The session to use for making this request.
        :param host: The host to migrate the server to. (Optional)
        :returns: None
        """
        if host and not utils.supports_microversion(session, '2.56'):
            raise ValueError(
                "The 'host' option is only supported on microversion 2.56 or "
                "greater."
            )

        body: dict[str, object] = {"migrate": None}
        if host:
            body["migrate"] = {"host": host}
        self._action(session, body)

    def trigger_crash_dump(self, session: adapter.Adapter) -> None:
        """Trigger a crash dump for the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        body = {"trigger_crash_dump": None}
        self._action(session, body)

    def get_console_output(
        self, session: adapter.Adapter, length: int | None = None
    ) -> dict[str, object]:
        """Get console output for the server.

        :param session: The session to use for making this request.
        :param length: The max length of the console output to return.
            (Optional)
        :returns: The console output.
        """
        body: dict[str, Any] = {"os-getConsoleOutput": {}}
        if length is not None:
            body["os-getConsoleOutput"]["length"] = length
        resp = self._action(session, body)
        return cast(dict[str, object], resp.json())

    def get_console_url(
        self, session: adapter.Adapter, console_type: str
    ) -> dict[str, object]:
        """Get the console URL for the server.

        :param session: The session to use for making this request.
        :param console_type: The type of console to return. This is
            cloud-specific. One of: ``novnc``, ``xvpvnc``, ``spice-html5``,
            ``spice-direct`` (after Nova microversion 2.99), ``rdp-html5``,
            or ``serial``.
        :returns: The console URL.
        """
        action = CONSOLE_TYPE_ACTION_MAPPING.get(console_type)
        if not action:
            raise ValueError(f"Unsupported console type {console_type}")
        body = {action: {'type': console_type}}
        resp = self._action(session, body)
        return cast(dict[str, object], resp.json()['console'])

    def live_migrate(
        self,
        session: adapter.Adapter,
        host: str | None,
        force: bool,
        block_migration: bool | str | None,
        disk_over_commit: bool | None = False,
    ) -> None:
        """Live migrate the server.

        :param session: The session to use for making this request.
        :param host: The host to live migrate the server to. (Optional)
        :param force: Whether to force the migration. (Optional)
        :param block_migration: Whether to do block migration. One of:
            ``True``, ``False``, ``'auto'``. (Optional)
        :param disk_over_commit: Whether to allow disk over-commit on the
            destination host. (Optional)
        :returns: None
        """
        if utils.supports_microversion(session, '2.30'):
            return self._live_migrate_30(
                session,
                host,
                force=force,
                block_migration=block_migration,
            )
        elif utils.supports_microversion(session, '2.25'):
            return self._live_migrate_25(
                session,
                host,
                force=force,
                block_migration=block_migration,
            )
        else:
            return self._live_migrate(
                session,
                host,
                force=force,
                block_migration=block_migration,
                disk_over_commit=disk_over_commit,
            )

    def _live_migrate_30(
        self,
        session: adapter.Adapter,
        host: str | None,
        force: bool,
        block_migration: bool | str | None,
    ) -> None:
        microversion = '2.30'
        body: dict[str, object] = {'host': None}
        if block_migration is None:
            block_migration = 'auto'
        body['block_migration'] = block_migration
        if host:
            body['host'] = host
            if force:
                body['force'] = force
        self._action(
            session,
            {'os-migrateLive': body},
            microversion=microversion,
        )

    def _live_migrate_25(
        self,
        session: adapter.Adapter,
        host: str | None,
        force: bool,
        block_migration: bool | str | None,
    ) -> None:
        microversion = '2.25'
        body: dict[str, object] = {'host': None}
        if block_migration is None:
            block_migration = 'auto'
        body['block_migration'] = block_migration
        if host:
            body['host'] = host
            if not force:
                raise ValueError(
                    "Live migration on this cloud implies 'force' "
                    "if the 'host' option has been given and it is not "
                    "possible to disable. It is recommended to not use 'host' "
                    "at all on this cloud as it is inherently unsafe, but if "
                    "it is unavoidable, please supply 'force=True' so that it "
                    "is clear you understand the risks."
                )
        self._action(
            session,
            {'os-migrateLive': body},
            microversion=microversion,
        )

    def _live_migrate(
        self,
        session: adapter.Adapter,
        host: str | None,
        force: bool,
        block_migration: bool | str | None,
        disk_over_commit: bool | None,
    ) -> None:
        microversion = None
        body: dict[str, object] = {
            'host': None,
        }
        if block_migration == 'auto':
            raise ValueError(
                "Live migration on this cloud does not support 'auto' as "
                "a parameter to block_migration, but only True and False."
            )
        body['block_migration'] = block_migration or False
        body['disk_over_commit'] = disk_over_commit or False
        if host:
            body['host'] = host
            if not force:
                raise ValueError(
                    "Live migration on this cloud implies 'force' "
                    "if the 'host' option has been given and it is not "
                    "possible to disable. It is recommended to not use 'host' "
                    "at all on this cloud as it is inherently unsafe, but if "
                    "it is unavoidable, please supply 'force=True' so that it "
                    "is clear you understand the risks."
                )
        self._action(
            session,
            {'os-migrateLive': body},
            microversion=microversion,
        )

    def fetch_topology(
        self, session: adapter.Adapter
    ) -> dict[str, object] | None:
        """Fetch the topology information for the server.

        :param session: The session to use for making this request.
        :returns: None
        """
        utils.require_microversion(session, '2.78')

        url = utils.urljoin(Server.base_path, self.id, 'topology')

        response = session.get(url)

        exceptions.raise_from_response(response)

        try:
            return cast(dict[str, object], response.json())
        except ValueError:
            return None

    def fetch_security_groups(self, session: adapter.Adapter) -> Self:
        """Fetch security groups of the server.

        :param session: The session to use for making this request.
        :returns: Updated Server instance.
        """
        url = utils.urljoin(Server.base_path, self.id, 'os-security-groups')

        response = session.get(url)

        exceptions.raise_from_response(response)

        try:
            data = response.json()
            if 'security_groups' in data:
                self.security_groups = data['security_groups']
        except ValueError:
            pass

        return self


ServerDetail = Server



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.container_infrastructure_management.v1 import cluster_template
from openstack.tests.unit import base

EXAMPLE = {
    "insecure_registry": None,
    "http_proxy": "http://10.164.177.169:8080",
    "updated_at": None,
    "floating_ip_enabled": True,
    "fixed_subnet": None,
    "master_flavor_id": None,
    "uuid": "085e1c4d-4f68-4bfd-8462-74b9e14e4f39",
    "no_proxy": "10.0.0.0/8,172.0.0.0/8,192.0.0.0/8,localhost",
    "https_proxy": "http://10.164.177.169:8080",
    "tls_disabled": False,
    "keypair_id": "kp",
    "public": False,
    "labels": {},
    "docker_volume_size": 3,
    "server_type": "vm",
    "external_network_id": "public",
    "cluster_distro": "fedora-atomic",
    "image_id": "fedora-atomic-latest",
    "volume_driver": "cinder",
    "registry_enabled": False,
    "docker_storage_driver": "devicemapper",
    "apiserver_port": None,
    "name": "k8s-bm2",
    "created_at": "2016-08-29T02:08:08+00:00",
    "network_driver": "flannel",
    "fixed_network": None,
    "coe": "kubernetes",
    "flavor_id": "m1.small",
    "master_lb_enabled": True,
    "dns_nameserver": "8.8.8.8",
    "hidden": True,
}


class TestClusterTemplate(base.TestCase):
    def test_basic(self):
        sot = cluster_template.ClusterTemplate()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('clustertemplates', sot.resources_key)
        self.assertEqual('/clustertemplates', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = cluster_template.ClusterTemplate(**EXAMPLE)

        self.assertEqual(EXAMPLE['apiserver_port'], sot.apiserver_port)
        self.assertEqual(EXAMPLE['cluster_distro'], sot.cluster_distro)
        self.assertEqual(EXAMPLE['coe'], sot.coe)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(
            EXAMPLE['docker_storage_driver'], sot.docker_storage_driver
        )
        self.assertEqual(EXAMPLE['docker_volume_size'], sot.docker_volume_size)
        self.assertEqual(EXAMPLE['dns_nameserver'], sot.dns_nameserver)
        self.assertEqual(
            EXAMPLE['external_network_id'], sot.external_network_id
        )
        self.assertEqual(EXAMPLE['fixed_network'], sot.fixed_network)
        self.assertEqual(EXAMPLE['fixed_subnet'], sot.fixed_subnet)
        self.assertEqual(EXAMPLE['flavor_id'], sot.flavor_id)
        self.assertEqual(EXAMPLE['http_proxy'], sot.http_proxy)
        self.assertEqual(EXAMPLE['https_proxy'], sot.https_proxy)
        self.assertEqual(EXAMPLE['image_id'], sot.image_id)
        self.assertEqual(EXAMPLE['insecure_registry'], sot.insecure_registry)
        self.assertEqual(
            EXAMPLE['floating_ip_enabled'], sot.is_floating_ip_enabled
        )
        self.assertEqual(EXAMPLE['hidden'], sot.is_hidden)
        self.assertEqual(
            EXAMPLE['master_lb_enabled'], sot.is_master_lb_enabled
        )
        self.assertEqual(EXAMPLE['tls_disabled'], sot.is_tls_disabled)
        self.assertEqual(EXAMPLE['public'], sot.is_public)
        self.assertEqual(EXAMPLE['registry_enabled'], sot.is_registry_enabled)
        self.assertEqual(EXAMPLE['keypair_id'], sot.keypair_id)
        self.assertEqual(EXAMPLE['master_flavor_id'], sot.master_flavor_id)
        self.assertEqual(EXAMPLE['network_driver'], sot.network_driver)
        self.assertEqual(EXAMPLE['no_proxy'], sot.no_proxy)
        self.assertEqual(EXAMPLE['server_type'], sot.server_type)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['uuid'], sot.uuid)
        self.assertEqual(EXAMPLE['volume_driver'], sot.volume_driver)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import tempfile
from unittest import mock

from openstack.cloud import _object_store
from openstack import exceptions
from openstack.object_store.v1 import _proxy
from openstack.object_store.v1 import container
from openstack.object_store.v1 import obj
from openstack.tests.unit import base
from openstack import utils


class BaseTestObject(base.TestCase):
    def setUp(self):
        super().setUp()

        self.container = self.getUniqueString()
        self.object = self.getUniqueString()
        self.endpoint = self.cloud.object_store.get_endpoint()
        self.container_endpoint = f'{self.endpoint}/{self.container}'
        self.object_endpoint = f'{self.container_endpoint}/{self.object}'

    def _compare_containers(self, exp, real):
        self.assertDictEqual(
            container.Container(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def _compare_objects(self, exp, real):
        self.assertDictEqual(
            obj.Object(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )


class TestObject(BaseTestObject):
    def test_create_container(self):
        """Test creating a (private) container"""
        self.register_uris(
            [
                dict(
                    method='HEAD', uri=self.container_endpoint, status_code=404
                ),
                dict(
                    method='PUT',
                    uri=self.container_endpoint,
                    status_code=201,
                    headers={
                        'Date': 'Fri, 16 Dec 2016 18:21:20 GMT',
                        'Content-Length': '0',
                        'Content-Type': 'text/html; charset=UTF-8',
                    },
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                ),
            ]
        )

        self.cloud.create_container(self.container)
        self.assert_calls()

    def test_create_container_public(self):
        """Test creating a public container"""
        self.register_uris(
            [
                dict(
                    method='HEAD', uri=self.container_endpoint, status_code=404
                ),
                dict(
                    method='PUT',
                    uri=self.container_endpoint,
                    status_code=201,
                    headers={
                        'Date': 'Fri, 16 Dec 2016 18:21:20 GMT',
                        'Content-Length': '0',
                        'Content-Type': 'text/html; charset=UTF-8',
                        'x-container-read': _object_store.OBJECT_CONTAINER_ACLS[  # noqa: E501
                            'public'
                        ],
                    },
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                ),
            ]
        )

        self.cloud.create_container(self.container, public=True)
        self.assert_calls()

    def test_create_container_exists(self):
        """Test creating a container that exists."""
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                )
            ]
        )

        container = self.cloud.create_container(self.container)

        self.assert_calls()
        self.assertIsNotNone(container)

    def test_delete_container(self):
        self.register_uris(
            [dict(method='DELETE', uri=self.container_endpoint)]
        )

        self.assertTrue(self.cloud.delete_container(self.container))
        self.assert_calls()

    def test_delete_container_404(self):
        """No exception when deleting a container that does not exist"""
        self.register_uris(
            [
                dict(
                    method='DELETE',
                    uri=self.container_endpoint,
                    status_code=404,
                )
            ]
        )

        self.assertFalse(self.cloud.delete_container(self.container))
        self.assert_calls()

    def test_delete_container_error(self):
        """Non-404 swift error re-raised as OSCE"""
        # 409 happens if the container is not empty
        self.register_uris(
            [
                dict(
                    method='DELETE',
                    uri=self.container_endpoint,
                    status_code=409,
                )
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.delete_container,
            self.container,
        )
        self.assert_calls()

    def test_update_container(self):
        headers = {
            'x-container-read': _object_store.OBJECT_CONTAINER_ACLS['public']
        }
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(headers=headers),
                )
            ]
        )

        self.cloud.update_container(self.container, headers)
        self.assert_calls()

    def test_update_container_error(self):
        """Swift error re-raised as OSCE"""
        # This test is of questionable value - the swift API docs do not
        # declare error codes (other than 404 for the container) for this
        # method, and I cannot make a synthetic failure to validate a real
        # error code. So we're really just testing the shade adapter error
        # raising logic here, rather than anything specific to swift.
        self.register_uris(
            [dict(method='POST', uri=self.container_endpoint, status_code=409)]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.update_container,
            self.container,
            dict(foo='bar'),
        )
        self.assert_calls()

    def test_set_container_access_public(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={
                            'x-container-read': _object_store.OBJECT_CONTAINER_ACLS[  # noqa: E501
                                'public'
                            ]
                        }
                    ),
                )
            ]
        )

        self.cloud.set_container_access(self.container, 'public')

        self.assert_calls()

    def test_set_container_access_private(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={
                            'x-container-read': _object_store.OBJECT_CONTAINER_ACLS[  # noqa: E501
                                'private'
                            ]
                        }
                    ),
                )
            ]
        )

        self.cloud.set_container_access(self.container, 'private')

        self.assert_calls()

    def test_set_container_access_invalid(self):
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.set_container_access,
            self.container,
            'invalid',
        )

    def test_get_container_access(self):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'x-container-read': str(
                            _object_store.OBJECT_CONTAINER_ACLS['public']
                        )
                    },
                )
            ]
        )
        access = self.cloud.get_container_access(self.container)
        self.assertEqual('public', access)

    def test_get_container_invalid(self):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={'x-container-read': 'invalid'},
                )
            ]
        )

        with self.assertRaises(
            exceptions.SDKException,
            msg="Could not determine container access for ACL: invalid",
        ):
            self.cloud.get_container_access(self.container)

    def test_get_container_access_not_found(self):
        self.register_uris(
            [dict(method='HEAD', uri=self.container_endpoint, status_code=404)]
        )
        with self.assertRaises(
            exceptions.SDKException,
            msg=f"Container not found: {self.container}",
        ):
            self.cloud.get_container_access(self.container)

    def test_list_containers(self):
        endpoint = f'{self.endpoint}/'
        containers = [{'count': 0, 'bytes': 0, 'name': self.container}]

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=endpoint,
                    complete_qs=True,
                    json=containers,
                )
            ]
        )

        ret = self.cloud.list_containers()

        self.assert_calls()
        for a, b in zip(containers, ret):
            self._compare_containers(a, b)

    def test_list_containers_exception(self):
        endpoint = f'{self.endpoint}/'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=endpoint,
                    complete_qs=True,
                    status_code=416,
                )
            ]
        )

        self.assertRaises(exceptions.SDKException, self.cloud.list_containers)
        self.assert_calls()

    @mock.patch.object(_proxy, '_get_expiration', return_value=13345)
    def test_generate_form_signature_container_key(self, mock_expiration):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'X-Container-Meta-Temp-Url-Key': 'amazingly-secure-key',  # noqa: E501
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                )
            ]
        )
        self.assertEqual(
            (13345, '60731fb66d46c97cdcb79b6154363179c500b9d9'),
            self.cloud.object_store.generate_form_signature(
                self.container,
                object_prefix='prefix/location',
                redirect_url='https://example.com/location',
                max_file_size=1024 * 1024 * 1024,
                max_upload_count=10,
                timeout=1000,
                temp_url_key=None,
            ),
        )
        self.assert_calls()

    @mock.patch.object(_proxy, '_get_expiration', return_value=13345)
    def test_generate_form_signature_account_key(self, mock_expiration):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                ),
                dict(
                    method='HEAD',
                    uri=self.endpoint + '/',
                    headers={
                        'X-Account-Meta-Temp-Url-Key': 'amazingly-secure-key'
                    },
                ),
            ]
        )
        self.assertEqual(
            (13345, '3cb9bc83d5a4136421bb2c1f58b963740566646f'),
            self.cloud.object_store.generate_form_signature(
                self.container,
                object_prefix='prefix/location',
                redirect_url='https://example.com/location',
                max_file_size=1024 * 1024 * 1024,
                max_upload_count=10,
                timeout=1000,
                temp_url_key=None,
            ),
        )
        self.assert_calls()

    @mock.patch.object(_proxy, '_get_expiration', return_value=13345)
    def test_generate_form_signature_key_argument(self, mock_expiration):
        self.assertEqual(
            (13345, '1c283a05c6628274b732212d9a885265e6f67b63'),
            self.cloud.object_store.generate_form_signature(
                self.container,
                object_prefix='prefix/location',
                redirect_url='https://example.com/location',
                max_file_size=1024 * 1024 * 1024,
                max_upload_count=10,
                timeout=1000,
                temp_url_key='amazingly-secure-key',
            ),
        )
        self.assert_calls()

    def test_generate_form_signature_no_key(self):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={
                        'Content-Length': '0',
                        'X-Container-Object-Count': '0',
                        'Accept-Ranges': 'bytes',
                        'X-Storage-Policy': 'Policy-0',
                        'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                        'X-Timestamp': '1481912480.41664',
                        'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                        'X-Container-Bytes-Used': '0',
                        'Content-Type': 'text/plain; charset=utf-8',
                    },
                ),
                dict(method='HEAD', uri=self.endpoint + '/', headers={}),
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.object_store.generate_form_signature,
            self.container,
            object_prefix='prefix/location',
            redirect_url='https://example.com/location',
            max_file_size=1024 * 1024 * 1024,
            max_upload_count=10,
            timeout=1000,
            temp_url_key=None,
        )
        self.assert_calls()

    def test_set_account_temp_url_key(self):
        key = 'super-secure-key'

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.endpoint + '/',
                    status_code=204,
                    validate=dict(
                        headers={'x-account-meta-temp-url-key': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.endpoint + '/',
                    headers={'x-account-meta-temp-url-key': key},
                ),
            ]
        )
        self.cloud.object_store.set_account_temp_url_key(key)
        self.assert_calls()

    def test_set_account_temp_url_key_secondary(self):
        key = 'super-secure-key'

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.endpoint + '/',
                    status_code=204,
                    validate=dict(
                        headers={'x-account-meta-temp-url-key-2': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.endpoint + '/',
                    headers={'x-account-meta-temp-url-key-2': key},
                ),
            ]
        )
        self.cloud.object_store.set_account_temp_url_key(key, secondary=True)
        self.assert_calls()

    def test_set_container_temp_url_key(self):
        key = 'super-secure-key'

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={'x-container-meta-temp-url-key': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={'x-container-meta-temp-url-key': key},
                ),
            ]
        )
        self.cloud.object_store.set_container_temp_url_key(self.container, key)
        self.assert_calls()

    def test_set_container_temp_url_key_secondary(self):
        key = 'super-secure-key'

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={'x-container-meta-temp-url-key-2': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={'x-container-meta-temp-url-key-2': key},
                ),
            ]
        )
        self.cloud.object_store.set_container_temp_url_key(
            self.container, key, secondary=True
        )
        self.assert_calls()

    def test_list_objects(self):
        endpoint = f'{self.container_endpoint}?format=json'

        objects = [
            {
                'bytes': 20304400896,
                'last_modified': '2016-12-15T13:34:13.650090',
                'hash': 'daaf9ed2106d09bba96cf193d866445e',
                'name': self.object,
                'content_type': 'application/octet-stream',
            }
        ]

        self.register_uris(
            [dict(method='GET', uri=endpoint, complete_qs=True, json=objects)]
        )

        ret = self.cloud.list_objects(self.container)

        self.assert_calls()
        for a, b in zip(objects, ret):
            self._compare_objects(a, b)

    def test_list_objects_with_prefix(self):
        endpoint = f'{self.container_endpoint}?format=json&prefix=test'

        objects = [
            {
                'bytes': 20304400896,
                'last_modified': '2016-12-15T13:34:13.650090',
                'hash': 'daaf9ed2106d09bba96cf193d866445e',
                'name': self.object,
                'content_type': 'application/octet-stream',
            }
        ]

        self.register_uris(
            [dict(method='GET', uri=endpoint, complete_qs=True, json=objects)]
        )

        ret = self.cloud.list_objects(self.container, prefix='test')

        self.assert_calls()
        for a, b in zip(objects, ret):
            self._compare_objects(a, b)

    def test_list_objects_exception(self):
        endpoint = f'{self.container_endpoint}?format=json'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=endpoint,
                    complete_qs=True,
                    status_code=416,
                )
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.list_objects,
            self.container,
        )
        self.assert_calls()

    def test_delete_object(self):
        self.register_uris(
            [
                dict(
                    method='HEAD',
                    uri=self.object_endpoint,
                    headers={'X-Object-Meta': 'foo'},
                ),
                dict(
                    method='DELETE', uri=self.object_endpoint, status_code=204
                ),
            ]
        )

        self.assertTrue(self.cloud.delete_object(self.container, self.object))

        self.assert_calls()

    def test_delete_object_not_found(self):
        self.register_uris(
            [dict(method='HEAD', uri=self.object_endpoint, status_code=404)]
        )

        self.assertFalse(self.cloud.delete_object(self.container, self.object))

        self.assert_calls()

    def test_get_object(self):
        headers = {
            'Content-Length': '20304400896',
            'Content-Type': 'application/octet-stream',
            'Accept-Ranges': 'bytes',
            'Last-Modified': 'Thu, 15 Dec 2016 13:34:14 GMT',
            'Etag': '"b5c454b44fbd5344793e3fb7e3850768"',
            'X-Timestamp': '1481808853.65009',
            'X-Trans-Id': 'tx68c2a2278f0c469bb6de1-005857ed80dfw1',
            'Date': 'Mon, 19 Dec 2016 14:24:00 GMT',
            'X-Static-Large-Object': 'True',
            'X-Object-Meta-Mtime': '1481513709.168512',
        }
        response_headers = {k.lower(): v for k, v in headers.items()}
        text = 'test body'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.object_endpoint,
                    headers={
                        'Content-Length': '20304400896',
                        'Content-Type': 'application/octet-stream',
                        'Accept-Ranges': 'bytes',
                        'Last-Modified': 'Thu, 15 Dec 2016 13:34:14 GMT',
                        'Etag': '"b5c454b44fbd5344793e3fb7e3850768"',
                        'X-Timestamp': '1481808853.65009',
                        'X-Trans-Id': 'tx68c2a2278f0c469bb6de1-005857ed80dfw1',
                        'Date': 'Mon, 19 Dec 2016 14:24:00 GMT',
                        'X-Static-Large-Object': 'True',
                        'X-Object-Meta-Mtime': '1481513709.168512',
                    },
                    text='test body',
                )
            ]
        )

        resp = self.cloud.get_object(self.container, self.object)

        self.assert_calls()

        self.assertEqual((response_headers, text), resp)

    def test_stream_object(self):
        text = b'test body'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.object_endpoint,
                    headers={
                        'Content-Length': '20304400896',
                        'Content-Type': 'application/octet-stream',
                        'Accept-Ranges': 'bytes',
                        'Last-Modified': 'Thu, 15 Dec 2016 13:34:14 GMT',
                        'Etag': '"b5c454b44fbd5344793e3fb7e3850768"',
                        'X-Timestamp': '1481808853.65009',
                        'X-Trans-Id': 'tx68c2a2278f0c469bb6de1-005857ed80dfw1',
                        'Date': 'Mon, 19 Dec 2016 14:24:00 GMT',
                        'X-Static-Large-Object': 'True',
                        'X-Object-Meta-Mtime': '1481513709.168512',
                    },
                    text='test body',
                )
            ]
        )

        response_text = b''
        for data in self.cloud.stream_object(self.container, self.object):
            response_text += data

        self.assert_calls()

        self.assertEqual(text, response_text)

    def test_stream_object_not_found(self):
        self.register_uris(
            [
                dict(method='GET', uri=self.object_endpoint, status_code=404),
            ]
        )

        response_text = b''
        for data in self.cloud.stream_object(self.container, self.object):
            response_text += data

        self.assert_calls()

        self.assertEqual(b'', response_text)

    def test_get_object_not_found(self):
        self.register_uris(
            [dict(method='GET', uri=self.object_endpoint, status_code=404)]
        )

        self.assertIsNone(self.cloud.get_object(self.container, self.object))

        self.assert_calls()

    def test_get_object_exception(self):
        self.register_uris(
            [dict(method='GET', uri=self.object_endpoint, status_code=416)]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.get_object,
            self.container,
            self.object,
        )

        self.assert_calls()

    def test_get_object_segment_size_below_min(self):
        # Register directly becuase we make multiple calls. The number
        # of calls we make isn't interesting - what we do with the return
        # values is. Don't run assert_calls for the same reason.
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': 1000},
                        slo={'min_segment_size': 500},
                    ),
                    headers={'Content-Type': 'application/json'},
                )
            ]
        )
        self.assertEqual(500, self.cloud.get_object_segment_size(400))
        self.assertEqual(900, self.cloud.get_object_segment_size(900))
        self.assertEqual(1000, self.cloud.get_object_segment_size(1000))
        self.assertEqual(1000, self.cloud.get_object_segment_size(1100))

    def test_get_object_segment_size_http_404(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    status_code=404,
                    reason='Not Found',
                )
            ]
        )
        self.assertEqual(
            _proxy.DEFAULT_OBJECT_SEGMENT_SIZE,
            self.cloud.get_object_segment_size(None),
        )
        self.assert_calls()

    def test_get_object_segment_size_http_412(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    status_code=412,
                    reason='Precondition failed',
                )
            ]
        )
        self.assertEqual(
            _proxy.DEFAULT_OBJECT_SEGMENT_SIZE,
            self.cloud.get_object_segment_size(None),
        )
        self.assert_calls()

    def test_update_container_cors(self):
        headers = {
            'X-Container-Meta-Web-Index': 'index.html',
            'X-Container-Meta-Access-Control-Allow-Origin': '*',
        }
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(headers=headers),
                )
            ]
        )
        self.cloud.update_container(self.container, headers=headers)
        self.assert_calls()


class TestObjectUploads(BaseTestObject):
    def setUp(self):
        super().setUp()

        self.content = self.getUniqueString().encode('latin-1')
        self.object_file = tempfile.NamedTemporaryFile(delete=False)
        self.object_file.write(self.content)
        self.object_file.close()
        self.md5, self.sha256 = utils._get_file_hashes(self.object_file.name)
        self.endpoint = self.cloud.object_store.get_endpoint()

    def test_create_object(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': 1000},
                        slo={'min_segment_size': 500},
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        }
                    ),
                ),
            ]
        )

        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
        )

        self.assert_calls()

    def test_create_object_index_rax(self):
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri='{endpoint}/{container}/{object}'.format(
                        endpoint=self.endpoint,
                        container=self.container,
                        object='index.html',
                    ),
                    status_code=201,
                    validate=dict(
                        headers={
                            'access-control-allow-origin': '*',
                            'content-type': 'text/html',
                        }
                    ),
                )
            ]
        )

        headers = {
            'access-control-allow-origin': '*',
            'content-type': 'text/html',
        }
        self.cloud.create_object(
            self.container, name='index.html', data='', **headers
        )

        self.assert_calls()

    def test_create_directory_marker_object(self):
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(
                        headers={
                            'content-type': 'application/directory',
                        }
                    ),
                )
            ]
        )

        self.cloud.create_directory_marker_object(
            container=self.container, name=self.object
        )

        self.assert_calls()

    def test_create_dynamic_large_object(self):
        max_file_size = 2
        min_file_size = 1

        uris_to_mock = [
            dict(
                method='GET',
                uri='https://object-store.example.com/info',
                json=dict(
                    swift={'max_file_size': max_file_size},
                    slo={'min_segment_size': min_file_size},
                ),
            ),
            dict(
                method='HEAD',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=404,
            ),
        ]

        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/{index:0>6}',
                    status_code=201,
                )
                for index, offset in enumerate(
                    range(0, len(self.content), max_file_size)
                )
            ]
        )

        uris_to_mock.append(
            dict(
                method='PUT',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=201,
                validate=dict(
                    headers={
                        'x-object-manifest': f'{self.container}/{self.object}',
                        'x-object-meta-x-sdk-md5': self.md5,
                        'x-object-meta-x-sdk-sha256': self.sha256,
                    }
                ),
            )
        )
        self.register_uris(uris_to_mock)
        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=False,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

        for key, value in self.calls[-1]['headers'].items():
            self.assertEqual(
                value,
                self.adapter.request_history[-1].headers[key],
                'header mismatch in manifest call',
            )

    def test_create_static_large_object(self):
        max_file_size = 25
        min_file_size = 1

        uris_to_mock = [
            dict(
                method='GET',
                uri='https://object-store.example.com/info',
                json=dict(
                    swift={'max_file_size': max_file_size},
                    slo={'min_segment_size': min_file_size},
                ),
            ),
            dict(
                method='HEAD',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=404,
            ),
        ]

        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/{index:0>6}',
                    status_code=201,
                    headers=dict(Etag=f'etag{index}'),
                )
                for index, offset in enumerate(
                    range(0, len(self.content), max_file_size)
                )
            ]
        )

        uris_to_mock.append(
            dict(
                method='PUT',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=201,
                validate=dict(
                    params={'multipart-manifest', 'put'},
                    headers={
                        'x-object-meta-x-sdk-md5': self.md5,
                        'x-object-meta-x-sdk-sha256': self.sha256,
                    },
                ),
            )
        )
        self.register_uris(uris_to_mock)

        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=True,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

        for key, value in self.calls[-1]['headers'].items():
            self.assertEqual(
                value,
                self.adapter.request_history[-1].headers[key],
                'header mismatch in manifest call',
            )

        base_object = f'/{self.container}/{self.object}'

        self.assertEqual(
            [
                {
                    'path': f"{base_object}/000000",
                    'size_bytes': 25,
                    'etag': 'etag0',
                },
                {
                    'path': f"{base_object}/000001",
                    'size_bytes': 25,
                    'etag': 'etag1',
                },
                {
                    'path': f"{base_object}/000002",
                    'size_bytes': 25,
                    'etag': 'etag2',
                },
                {
                    'path': f"{base_object}/000003",
                    'size_bytes': len(self.object) - 75,
                    'etag': 'etag3',
                },
            ],
            self.adapter.request_history[-1].json(),
        )

    def test_slo_manifest_retry(self):
        """
        Uploading the SLO manifest file should be retried up to 3 times before
        giving up. This test should succeed on the 3rd and final attempt.
        """
        max_file_size = 25
        min_file_size = 1

        uris_to_mock = [
            dict(
                method='GET',
                uri='https://object-store.example.com/info',
                json=dict(
                    swift={'max_file_size': max_file_size},
                    slo={'min_segment_size': min_file_size},
                ),
            ),
            dict(
                method='HEAD',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=404,
            ),
        ]

        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/{index:0>6}',
                    status_code=201,
                    headers=dict(Etag=f'etag{index}'),
                )
                for index, offset in enumerate(
                    range(0, len(self.content), max_file_size)
                )
            ]
        )

        # manifest file upload calls
        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=400,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=400,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
            ]
        )

        self.register_uris(uris_to_mock)

        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=True,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

        for key, value in self.calls[-1]['headers'].items():
            self.assertEqual(
                value,
                self.adapter.request_history[-1].headers[key],
                'header mismatch in manifest call',
            )

        base_object = f'/{self.container}/{self.object}'

        self.assertEqual(
            [
                {
                    'path': f"{base_object}/000000",
                    'size_bytes': 25,
                    'etag': 'etag0',
                },
                {
                    'path': f"{base_object}/000001",
                    'size_bytes': 25,
                    'etag': 'etag1',
                },
                {
                    'path': f"{base_object}/000002",
                    'size_bytes': 25,
                    'etag': 'etag2',
                },
                {
                    'path': f"{base_object}/000003",
                    'size_bytes': len(self.object) - 75,
                    'etag': 'etag3',
                },
            ],
            self.adapter.request_history[-1].json(),
        )

    def test_slo_manifest_fail(self):
        """
        Uploading the SLO manifest file should be retried up to 3 times before
        giving up. This test fails all 3 attempts and should verify that we
        delete uploaded segments that begin with the object prefix.
        """
        max_file_size = 25
        min_file_size = 1

        uris_to_mock = [
            dict(
                method='GET',
                uri='https://object-store.example.com/info',
                json=dict(
                    swift={'max_file_size': max_file_size},
                    slo={'min_segment_size': min_file_size},
                ),
            ),
            dict(
                method='HEAD',
                uri=f'{self.endpoint}/{self.container}/{self.object}',
                status_code=404,
            ),
        ]

        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/{index:0>6}',
                    status_code=201,
                    headers=dict(Etag=f'etag{index}'),
                )
                for index, offset in enumerate(
                    range(0, len(self.content), max_file_size)
                )
            ]
        )

        # manifest file upload calls
        uris_to_mock.extend(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=400,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=400,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=400,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
            ]
        )

        # Cleaning up image upload segments involves calling the
        # delete_autocreated_image_objects() API method which will list
        # objects (LIST), get the object metadata (HEAD), then delete the
        # object (DELETE).
        uris_to_mock.extend(
            [
                dict(
                    method='GET',
                    uri=f'{self.endpoint}/images?format=json&prefix={self.object}',
                    complete_qs=True,
                    json=[
                        {
                            'content_type': 'application/octet-stream',
                            'bytes': 1437258240,
                            'hash': '249219347276c331b87bf1ac2152d9af',
                            'last_modified': '2015-02-16T17:50:05.289600',
                            'name': self.object,
                        }
                    ],
                ),
                dict(
                    method='HEAD',
                    uri=f'{self.endpoint}/images/{self.object}',
                    headers={
                        'X-Timestamp': '1429036140.50253',
                        'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                        'Content-Length': '1290170880',
                        'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                        'X-Object-Meta-x-sdk-autocreated': 'true',
                        'X-Object-Meta-X-Shade-Sha256': 'does not matter',
                        'X-Object-Meta-X-Shade-Md5': 'does not matter',
                        'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                        'Accept-Ranges': 'bytes',
                        'X-Static-Large-Object': 'false',
                        'Content-Type': 'application/octet-stream',
                        'Etag': '249219347276c331b87bf1ac2152d9af',
                    },
                ),
                dict(
                    method='DELETE',
                    uri=f'{self.endpoint}/images/{self.object}',
                ),
            ]
        )

        self.register_uris(uris_to_mock)

        # image_api_use_tasks needs to be set to True in order for the API
        # method delete_autocreated_image_objects() to do the cleanup.
        self.cloud.image_api_use_tasks = True

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.create_object,
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=True,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

    def test_object_segment_retry_failure(self):
        max_file_size = 25
        min_file_size = 1

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': max_file_size},
                        slo={'min_segment_size': min_file_size},
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000000',
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000001',
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000002',
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000003',
                    status_code=501,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                ),
            ]
        )

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.create_object,
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=True,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

    def test_object_segment_retries(self):
        max_file_size = 25
        min_file_size = 1

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': max_file_size},
                        slo={'min_segment_size': min_file_size},
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=404,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000000',
                    headers={'etag': 'etag0'},
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000001',
                    headers={'etag': 'etag1'},
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000002',
                    headers={'etag': 'etag2'},
                    status_code=201,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000003',
                    status_code=501,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}/000003',
                    status_code=201,
                    headers={'etag': 'etag3'},
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(
                        params={'multipart-manifest', 'put'},
                        headers={
                            'x-object-meta-x-sdk-md5': self.md5,
                            'x-object-meta-x-sdk-sha256': self.sha256,
                        },
                    ),
                ),
            ]
        )

        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            use_slo=True,
        )

        # After call 3, order become indeterminate because of thread pool
        self.assert_calls(stop_after=3)

        for key, value in self.calls[-1]['headers'].items():
            self.assertEqual(
                value,
                self.adapter.request_history[-1].headers[key],
                'header mismatch in manifest call',
            )

        base_object = f'/{self.container}/{self.object}'

        self.assertEqual(
            [
                {
                    'path': f"{base_object}/000000",
                    'size_bytes': 25,
                    'etag': 'etag0',
                },
                {
                    'path': f"{base_object}/000001",
                    'size_bytes': 25,
                    'etag': 'etag1',
                },
                {
                    'path': f"{base_object}/000002",
                    'size_bytes': 25,
                    'etag': 'etag2',
                },
                {
                    'path': f"{base_object}/000003",
                    'size_bytes': len(self.object) - 75,
                    'etag': 'etag3',
                },
            ],
            self.adapter.request_history[-1].json(),
        )

    def test_create_object_skip_checksum(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://object-store.example.com/info',
                    json=dict(
                        swift={'max_file_size': 1000},
                        slo={'min_segment_size': 500},
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=200,
                ),
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(headers={}),
                ),
            ]
        )

        self.cloud.create_object(
            container=self.container,
            name=self.object,
            filename=self.object_file.name,
            generate_checksums=False,
        )

        self.assert_calls()

    def test_create_object_data(self):
        self.register_uris(
            [
                dict(
                    method='PUT',
                    uri=f'{self.endpoint}/{self.container}/{self.object}',
                    status_code=201,
                    validate=dict(
                        headers={},
                        data=self.content,
                    ),
                ),
            ]
        )

        self.cloud.create_object(
            container=self.container, name=self.object, data=self.content
        )

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.cloud import inventory
import openstack.config
from openstack.tests.unit import base
from openstack.tests.unit.cloud import fakes


class TestInventory(base.TestCase):
    def setUp(self):
        super().setUp()

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test__init(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        mock_config.assert_called_once_with(
            config_files=openstack.config.loader.CONFIG_FILES
        )
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        self.assertTrue(mock_config.return_value.get_all.called)

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test__init_one_cloud(self, mock_cloud, mock_config):
        mock_config.return_value.get_one.return_value = [{}]

        inv = inventory.OpenStackInventory(cloud='supercloud')

        mock_config.assert_called_once_with(
            config_files=openstack.config.loader.CONFIG_FILES
        )
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        self.assertFalse(mock_config.return_value.get_all.called)
        mock_config.return_value.get_one.assert_called_once_with('supercloud')

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test_list_hosts(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        server = dict(id='server_id', name='server_name')
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        inv.clouds[0].list_servers.return_value = [server]
        inv.clouds[0].get_openstack_vars.return_value = server

        ret = inv.list_hosts()

        inv.clouds[0].list_servers.assert_called_once_with(
            detailed=True, all_projects=False
        )
        self.assertFalse(inv.clouds[0].get_openstack_vars.called)
        self.assertEqual([server], ret)

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test_list_hosts_no_detail(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        server = self.cloud._normalize_server(
            fakes.make_fake_server('1234', 'test', 'ACTIVE', addresses={})
        )
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        inv.clouds[0].list_servers.return_value = [server]

        inv.list_hosts(expand=False)

        inv.clouds[0].list_servers.assert_called_once_with(
            detailed=False, all_projects=False
        )
        self.assertFalse(inv.clouds[0].get_openstack_vars.called)

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test_list_hosts_all_projects(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        server = dict(id='server_id', name='server_name')
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        inv.clouds[0].list_servers.return_value = [server]
        inv.clouds[0].get_openstack_vars.return_value = server

        ret = inv.list_hosts(all_projects=True)

        inv.clouds[0].list_servers.assert_called_once_with(
            detailed=True, all_projects=True
        )
        self.assertFalse(inv.clouds[0].get_openstack_vars.called)
        self.assertEqual([server], ret)

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test_search_hosts(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        server = dict(id='server_id', name='server_name')
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        inv.clouds[0].list_servers.return_value = [server]
        inv.clouds[0].get_openstack_vars.return_value = server

        ret = inv.search_hosts('server_id')
        self.assertEqual([server], ret)

    @mock.patch("openstack.config.loader.OpenStackConfig")
    @mock.patch("openstack.connection.Connection")
    def test_get_host(self, mock_cloud, mock_config):
        mock_config.return_value.get_all.return_value = [{}]

        inv = inventory.OpenStackInventory()

        server = dict(id='server_id', name='server_name')
        self.assertIsInstance(inv.clouds, list)
        self.assertEqual(1, len(inv.clouds))
        inv.clouds[0].list_servers.return_value = [server]
        inv.clouds[0].get_openstack_vars.return_value = server

        ret = inv.get_host('server_id')
        self.assertEqual(server, ret)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack import exceptions
from openstack.identity.v3 import token
from openstack.tests.unit import base

IDENTIFIER = 'IDENTIFIER'
TOKEN_DATA = {
    'audit_ids': ['VcxU2JEMTjufVx7sVk7bPw'],
    'catalog': [
        {
            'endpoints': [
                {
                    'id': '068d1b359ee84b438266cb736d81de97',
                    'interface': 'public',
                    'region': 'RegionOne',
                    'region_id': 'RegionOne',
                    'url': 'http://example.com/v2.1',
                }
            ],
            'id': '050726f278654128aba89757ae25950c',
            'name': 'nova',
            'type': 'compute',
        }
    ],
    'domain': {'id': 'default', 'name': 'Default'},
    'expires_at': '2013-02-27T18:30:59.999999Z',
    'issued_at': '2013-02-27T16:30:59.999999Z',
    'methods': ['password'],
    'project': {
        'domain': {'id': 'default', 'name': 'Default'},
        'id': '8538a3f13f9541b28c2620eb19065e45',
        'name': 'admin',
    },
    'roles': [{'id': 'c703057be878458588961ce9a0ce686b', 'name': 'admin'}],
    'system': {'all': True},
    'user': {
        'domain': {'id': 'default', 'name': 'Default'},
        'id': '10a2e6e717a245d9acad3e5f97aeca3d',
        'name': 'admin',
        'password_expires_at': None,
    },
    'is_domain': False,
}

EXAMPLE = {'token': TOKEN_DATA}


class TestToken(base.TestCase):
    def setUp(self):
        super().setUp()
        self.session = mock.Mock(spec=adapter.Adapter)

    def test_basic(self):
        sot = token.Token()
        self.assertEqual('token', sot.resource_key)
        self.assertEqual('/auth/tokens', sot.base_path)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)
        self.assertFalse(sot.allow_head)

    def test_make_it(self):
        sot = token.Token(**TOKEN_DATA)
        self.assertEqual(TOKEN_DATA['audit_ids'], sot.audit_ids)
        self.assertEqual(TOKEN_DATA['catalog'], sot.catalog)
        self.assertEqual(TOKEN_DATA['expires_at'], sot.expires_at)
        self.assertEqual(TOKEN_DATA['issued_at'], sot.issued_at)
        self.assertEqual(TOKEN_DATA['methods'], sot.methods)
        self.assertEqual(TOKEN_DATA['user'], sot.user)
        self.assertEqual(TOKEN_DATA['project'], sot.project)
        self.assertEqual(TOKEN_DATA['domain'], sot.domain)
        self.assertEqual(TOKEN_DATA['is_domain'], sot.is_domain)
        self.assertEqual(TOKEN_DATA['system'], sot.system)
        self.assertEqual(TOKEN_DATA['roles'], sot.roles)

    def test_validate(self):
        response = mock.Mock()
        response.status_code = 200
        response.json.return_value = EXAMPLE
        response.headers = {'content-type': 'application/json'}
        self.session.get.return_value = response

        result = token.Token.validate(self.session, 'token')

        self.session.get.assert_called_once_with(
            '/auth/tokens', headers={'x-subject-token': 'token'}, params={}
        )
        self.assertIsInstance(result, token.Token)

    def test_validate_with_params(self):
        response = mock.Mock()
        response.status_code = 200
        response.json.return_value = EXAMPLE
        response.headers = {'content-type': 'application/json'}
        self.session.get.return_value = response

        result = token.Token.validate(
            self.session, 'token', nocatalog=True, allow_expired=True
        )

        self.session.get.assert_called_once_with(
            '/auth/tokens',
            headers={'x-subject-token': 'token'},
            params={'nocatalog': True, 'allow_expired': True},
        )
        self.assertIsInstance(result, token.Token)

    def test_validate_error(self):
        response = mock.Mock()
        response.status_code = 404
        response.json.return_value = {}
        response.headers = {'content-type': 'application/json'}
        self.session.get.return_value = response

        self.assertRaises(
            exceptions.NotFoundException,
            token.Token.validate,
            self.session,
            'token',
        )

    def test_check(self):
        response = mock.Mock()
        response.status_code = 200
        self.session.head.return_value = response

        result = token.Token.check(self.session, 'token')

        self.session.head.assert_called_once_with(
            '/auth/tokens', headers={'x-subject-token': 'token'}, params={}
        )
        self.assertTrue(result)

    def test_check_with_param(self):
        response = mock.Mock()
        response.status_code = 200
        self.session.head.return_value = response

        result = token.Token.check(self.session, 'token', allow_expired=True)

        self.session.head.assert_called_once_with(
            '/auth/tokens',
            headers={'x-subject-token': 'token'},
            params={'allow_expired': True},
        )
        self.assertTrue(result)

    def test_check_invalid_token(self):
        response = mock.Mock()
        response.status_code = 404
        self.session.head.return_value = response

        result = token.Token.check(self.session, 'token')

        self.session.head.assert_called_once_with(
            '/auth/tokens', headers={'x-subject-token': 'token'}, params={}
        )
        self.assertFalse(result)

    def test_revoke(self):
        response = mock.Mock()
        response.status_code = 204
        self.session.delete.return_value = response

        token.Token.revoke(self.session, 'token')

        self.session.delete.assert_called_once_with(
            '/auth/tokens', headers={'x-subject-token': 'token'}
        )

    def test_revoke_error(self):
        response = mock.Mock()
        response.status_code = 404
        response.json.return_value = {}
        response.headers = {'content-type': 'application/json'}
        self.session.delete.return_value = response

        self.assertRaises(
            exceptions.NotFoundException,
            token.Token.revoke,
            self.session,
            'token',
        )
